from . import (json_serialization, dill_serialization,
               pickle_serialization, cloudpickle, raw)

from .. import utils

formats = {}
def register_serialization(name, serializer, deserializer):
    formats[name] = (serializer, deserializer)

def register_default_serialization():
    register_serialization('json',
                           json_serialization.serialize,
                           json_serialization.deserialize)
    register_serialization('dill',
                           dill_serialization.serialize,
                           dill_serialization.deserialize)
    register_serialization('pickle',
                           pickle_serialization.serialize,
                           pickle_serialization.deserialize)
    register_serialization('cloudpickle',
                           cloudpickle.dumps,
                           pickle_serialization.deserialize)
    register_serialization('raw',
                           raw.serialize,
                           raw.deserialize)

register_default_serialization()

def serializer(fmt):
    return formats[fmt][0]

def deserializer(fmt):
    return formats[fmt][1]

"""
we're sending function call metadata as json, and
the rest as whatever format the user wants to send
separator splits it out
metadata:
fmt, auth_string, serialized_function, async, queue_name

data:
func_string, args_string, kwargs_string
"""
separator = '""""'
"""
format:
   msg_format : json object, serialized as json
       contains fmt : a list of formats for the remaining data objects
       and len : an array of the lengths of the data in the remaining data objects
   then the data block begins, which is N data strings
   encoded by the 'fmt'
"""
def pack_msg(*data, **kwargs):
    fmts = kwargs.pop('fmt')
    data_strings = []
    lengths = []
    for f, d in zip(fmts, data):
        data_string = serializer(f)(d)
        data_strings.append(data_string)
        lengths.append(len(data_string))
    msg_format = {'len' : lengths, 'fmt' : fmts}
    msg_format_string = serializer('json')(msg_format)
    return msg_format_string + separator + "".join(data_strings)

def unpack_msg(input_string, override_fmt=None):
    msg_format = unpack_msg_format(input_string)
    data = unpack_msg_data(msg_format, input_string, override_fmt=override_fmt)
    return msg_format, data

def unpack_msg_format(input_string):
    return deserializer('json')(input_string.split(separator, 1)[0])

def unpack_msg_data(msg_format, input_string, index=None, override_fmt=None):
    """won't work if you specify incorrect override_fmt, but
    override_fmt is used to make sure we don't use something like dill
    when we don't want it (Security)
    """
    _, data_string = input_string.split(separator, 1)
    def unpack_data(index):
        if override_fmt is None:
            fmt = msg_format['fmt'][index]
        else:
            fmt = override_fmt
        length = msg_format['len'][index]
        idx = sum(msg_format['len'][:index])
        func = deserializer(fmt)
        data = func(data_string[idx:idx+length])
        return data
    if index is not None:
        return unpack_data(index)
    else:
        return [unpack_data(x) for x in range(len(msg_format['fmt']))]

"""
RPC Call format:
msg_format: similar to msg above
metadata: contains func_string (optional)
result_fmt (mandatory)
queue_name (mandatory)
auth_string (optional)
async (mandatory)
data block : series of serialized dictionaries with
  func, args, kwargs

either the data_block must contain func, or the metadata must
contain func_string

we only want one data object (which is a dictionary with those things)
but we allow for passing in several, so that we can modify the
data block without deserializing the data block. This is because
un-dilling or un-pickling can be dangerous

as a result, all data blocks are converted into dicts, and then
updated, so that the last dicts values takes precedence
"""

def pack_rpc_call(metadata, *data, **kwargs):
    fmt = kwargs.pop('fmt', 'dill')
    #always use json for rpc metadata
    fmts = ['json'] + [fmt for d in data]
    return pack_msg(metadata, *data, fmt=fmts)

def append_rpc_data(input_string, data, fmt=None):
    msg_format, data_string = input_string.split(separator, 1)
    msg_format = deserializer('json')(msg_format)
    new_string = serializer(fmt)(data)
    msg_format['len'].append(len(new_string))
    msg_format['fmt'].append(fmt)
    return serializer('json')(msg_format) + separator + data_string + new_string

def unpack_rpc_call(input_string):
    msg_format, data = unpack_msg(input_string)
    metadata = data[0]
    data = utils.update_dictionaries(*data[1:])
    return msg_format, metadata, data

def unpack_rpc_metadata(input_string):
    msg_format = unpack_msg_format(input_string)
    return unpack_msg_data(msg_format, input_string, index=0, override_fmt='json')


"""
RPC result format (same as our message format, except only 1 data payload)
"""

def pack_result(metadata, data, fmt=None):
    return pack_msg(metadata, data, fmt=['json', fmt])

def unpack_result(input_string):
    return unpack_msg(input_string)

def pack_results(metadata_data_pairs, fmt=None):
    data = reduce(lambda x, y : x + y, metadata_data_pairs)
    fmt = [('json', f) for f in fmt]
    fmt = reduce(lambda x, y : x + y, fmt)
    return pack_msg(*data, fmt=fmt)

def unpack_results(input_string):
    msg_format, metadata_data_pairs = unpack_msg(input_string)
    metadata_data_pairs = [(metadata_data_pairs[x], metadata_data_pairs[x+1]) for x in \
                           range(0, len(metadata_data_pairs), 2)]
    return metadata_data_pairs



from os.path import join, exists, isdir, relpath, abspath, dirname
import datetime as dt
import posixpath
import logging
import tempfile
from os import stat, makedirs, remove
import random
import uuid
from cStringIO import StringIO
import time

from six import string_types
try:
    import gevent
except:
    gevent = None
from ..clients.http import Client
from .. import settings
from ..serialization import deserializer, serializer
from ..errors import KitchenSinkError
from ..utils.pathutils import urlsplit, dirsplit, urljoin

logger = logging.getLogger(__name__)

def hosts():
    if settings.prefix:
        host_key = settings.prefix + ":" + "hosts"
        hostinfo_key = settings.prefix + ":" + "hostinfo"
    else:
        host_key = "hosts"
        hostinfo_key = "hostinfo"
    possible_hosts = settings.redis_conn.smembers(host_key)
    hosts = settings.redis_conn.mget(*possible_hosts)
    hosts = [x for x in hosts if x is not None]
    return hosts

def _write(finput, f):
    if isinstance(finput, string_types):
        finput = StringIO(finput)
    if isinstance(f, string_types):
        f = open(f, 'wb+')
    try:
        while True:
            data = finput.read(settings.chunk_size)
            if data:
                f.write(data)
                if gevent:
                    gevent.sleep(0)
            else:
                break
    finally:
        finput.close()
        f.close()

def _read(path):
    with open(path, "rb") as f:
        return f.read()

class Catalog(object):
    def __init__(self, connection, datadir, host_url, prefix=None):
        """connection - redis connection
        host_url - url of current controller/host
        """
        if prefix is None:
            prefix = settings.prefix
        self.prefix = ""
        self.conn = connection
        self.datadir = datadir
        self.host_url = host_url

    def path_key(self, key):
        """redis key for storing local paths for remote data sources
        """
        if self.prefix:
            return self.prefix + ":data:path:" + key
        else:
            return "data:path:" + key

    def starting_key(self, key):
        """redis key for storing information about
        beginning to write a remote data source
        """
        if self.prefix:
            return self.prefix + ":datastarting:path:" + key
        else:
            return "datastarting:path:" + key

    def data_key(self, key):
        """redis key for metadata about the remote data source
        (filesizes, datatype, format)
        """
        if self.prefix:
            return self.prefix + ":data:info:" + key
        else:
            return "data:info:" + key

    def search(self, pattern):
        """search the catalog for remote data urls
        currently implemented with redis.keys which
        isn't recommended for production use
        """
        if self.prefix:
            prefix = self.prefix + ":data:info:"
        else:
            prefix = self.prefix + "data:info:"
        pattern = prefix + pattern
        ### implement this with scan later
        keys = self.conn.keys(pattern)
        return [x[len(prefix):] for x in keys]

    def setup_file_path_from_url(self, url):
        """given a url, setup the file path in the data directory
        ensuring that necessary subdirs are created, and that the
        file path is valid
        """
        splits = urlsplit(url, "")
        file_path = abspath(join(self.datadir, *splits))
        if not file_path.startswith(self.datadir):
            raise KitchenSinkError, "Must be inside datadir"
        if not exists(dirname(file_path)):
            makedirs(dirname(file_path))
        return file_path

    def write(self, finput, url, is_new=True, data_type="object", fmt="cloudpickle"):
        """writes a file, to a data url.
        is_new - is this a new object in a catalog, or a copy of an existing one.
        data_type - data_type of the object (object or file)
        fmt - one of our serialization formats
        if the targetfile path exists - do nothing (this should't happen, maybe we should throw
        an error)
        if this file is new, then write the metadata into the catalog
        otherwise just write the data to local disk
        """
        file_path = self.setup_file_path_from_url(url)
        if is_new:
            if self.url_exists(url):
                raise KitchenSinkError("path already being used")
            else:
                self.init_addition(file_path, url)
        if not exists(file_path):
            _write(finput, file_path)
        if is_new:
            size = stat(file_path).st_size
            self.set_metadata(url, size, data_type=data_type, fmt=fmt)
        self.add(file_path, url)
        return file_path

    def delete(self, url):
        """delete the data url from this node only.  to truly remove it
        you need to delete it from all nodes
        """
        start_key = self.starting_key(url)
        path_key = self.path_key(url)
        data_key = self.data_key(url)
        self.conn.hdel(path_key, self.host_url)
        self.conn.hdel(start_key, self.host_url)
        if not self.conn.exists(path_key):
            self.conn.delete(data_key)
        file_path = self.setup_file_path_from_url(url)
        remove(file_path)

    def write_chunked(self, iterator, url, is_new=True):
        """same as chunked, but write data from an iterator
        (we use chunked reads during pipelining data, so we can
        stream data in as it's being processed)
        """
        file_path = self.setup_file_path_from_url(url)
        if is_new:
            if self.url_exists(url):
                raise KitchenSinkError("path already being used")
            else:
                self.init_addition(file_path, url)
        with open(file_path, "wb+") as f:
            for chunk in iterator:
                f.write(chunk)
        if is_new:
            size = stat(file_path).st_size
            self.set_metadata(url, size, data_type=data_type)
        self.add(file_path, url)
        return file_path

    def url_exists(self, url):
        start_key = self.starting_key(url)
        path_key = self.path_key(url)
        data_key = self.data_key(url)
        return self.conn.exists(start_key) or \
            self.conn.exists(path_key) or \
            self.conn.exists(data_key)

    def add(self, file_path, url):
        """add a url to the catalog
        """
        path_key = self.path_key(url)
        self.conn.hset(path_key, self.host_url, file_path)
        return file_path

    def init_addition(self, file_path, url):
        """declare our intention that we are beginnning to add
        a data url to the catalog
        """
        start_key = self.starting_key(url)
        self.conn.hset(start_key, self.host_url, file_path)

    def set_metadata(self, url, size, data_type="object", fmt="cloudpickle"):
        data_key = self.data_key(url)
        self.conn.hmset(data_key, {'size' : str(size),
                                   'data_type' : data_type,
                                   'fmt' : fmt}
        )

    def get_chunked_iterator(self, url, length, host_url=None):
        """return a chunked iterator for the contents of a file
        at another host.  This is used for pipelining, so that
        we can stream data in while it's being written
        """
        if host_url is None:
            host_url = self.host_url
        data_read = 0
        c = Client(host_url, rpc_name='data', queue_name='data')
        while True:
            if data_read == length:
                break
            data = c._get_data(url, data_read, settings.chunk_size)
            if data.status_code != 200:
                raise KitchenSinkError("http error")
            data = data.raw.read()
            data_read += len(data)
            logger.debug ("read %s of %s from %s to %s" % (data_read, length,
                                                          host_url, settings.host_url))
            if gevent:
                gevent.sleep(0)
            if len(data) == 0:
                time.sleep(1.0)
            yield data

    def get(self, url, host=None):
        """returns a stream for the given data url
        it is up to the caller of this to close the stream!
        """
        hosts_info, data_info = self.get_info(url)
        if self.host_url in hosts_info:
            return open(hosts_info[self.host_url], 'rb')
        else:
            host = hosts_info.keys()[0]
            logger.info("retrieving %s from %s", url, host)
            c = Client(host, rpc_name='data', queue_name='data')
            return c._get_data(url).raw

    def get_info(self, url):
        hosts_info = self.conn.hgetall(self.path_key(url))
        data_info = self.conn.hgetall(self.data_key(url))
        if 'size' in data_info:
            data_info['size'] = int(data_info['size'])
        return (hosts_info, data_info)

    def get_file_path(self, url, unfinished=False):
        """retrieve file path for the url.
        unfinished means the data is not in the catalog yet
        (we're pipelining it)
        if the url does not exist on this host, return None
        """
        if unfinished:
            path = self.setup_file_path_from_url(url)
            if not exists(path):
                return None
        hosts_info, data_info = self.get_info(url)
        try:
            file_path = hosts_info[self.host_url]
        except KeyError:
            return None
        if not file_path.startswith(self.datadir):
            raise KitchenSinkError, "Must be inside datadir"
        return file_path

## should factor this out into a separate module
class RemoteData(object):
    """RemoteData objects can contain raw data (in memory, in self._raw,
    or on disk, in self._local_path.  Or as a deserialized object, in
    self._obj.

    Maybe this isn't the greatest approach, mixing them all together in one object
    but that's what we have for now.  Currently, if you ask for self.obj(), or
    self.local_path(), or self.raw(), we will retrieve it from what we perceive
    to be the cheapest source.  For example, if you have the stream in memory
    (self._raw), and you ask for the local path, we will write it to a temp file
    and return you that path
    """
    def __init__(self, obj=None, local_path=None, data_url=None,
                 raw=None,
                 rpc_url=None,
                 fmt="cloudpickle"):
        if rpc_url is None:
            rpc_url = settings.data_rpc_url
        self.rpc_url = rpc_url
        self.data_url = data_url
        self.fmt = fmt
        self._obj = obj
        self._local_path = local_path
        self._raw = raw

    def __setstate__(self, obj):
        self.data_url = obj['data_url']
        self.fmt = obj['fmt']
        self.rpc_url = settings.data_rpc_url
        self._obj = None
        self._local_path = None
        self._raw = None

    def __getstate__(self):
        return dict(data_url=self.data_url,
                    fmt=self.fmt)

    def client(self, rpc_url=None):
        if rpc_url is None:
            rpc_url = self.rpc_url
        return Client(rpc_url, rpc_name='data', queue_name='data')

    def _get_stream(self):
        """returns either the stream associated with a data_url
        """
        if settings.catalog:
            return settings.catalog.get(self.data_url)
        c = self.client(self.rpc_url)
        url = c.pick_host(self.data_url)
        resp = self.client(url)._get_data(self.data_url)
        return resp.raw

    def _put(self, f, data_type='object', fmt="cloudpickle"):
        logger.debug("posting %s to %s", self.data_url, self.rpc_url)
        f.seek(0)
        if settings.catalog:
            settings.catalog.write(f, self.data_url, is_new=True, data_type=data_type,
                                   fmt=self.fmt)
        else:
            c = self.client(self.rpc_url)
            return c._put_data(self.data_url, f, data_type=data_type, fmt=fmt)

    def _existing_file_path(self):
        if settings.catalog:
            name = settings.catalog.get_file_path(self.data_url)
            if name is not None:
                self._local_path = name
                return name

    def local_path(self):
        """provides a path to a local file that contains the
        contents of this remote data source (downloads if necessary)
        """
        if self._local_path:
            return self._local_path
        path = self._existing_file_path()
        if path is not None:
            return path

        ### if we have the data in memory, write it to a local file and return it
        if self._raw:
            name = tempfile.NamedTemporaryFile(prefix="ks-data-").name
            _write(self._raw, name)
            self._local_path = name
            return name
        ### if we have an in memory object, serialize it, write to a file
        ### and return it
        if self._obj is not None:
            name = tempfile.NamedTemporaryFile(prefix="ks-data-").name
            data = serializer(self.fmt)(self._obj)
            _write(data, name)
            self._local_path = name
            return name
        ### grab the stream, write to temporary file, return the path
        stream = self._get_stream()
        try:
            name = tempfile.NamedTemporaryFile(prefix="ks-data-").name
            _write(stream, name)
            self._local_path = name
            return name
        finally:
            stream.close()

    def raw(self):
        """provides raw contents of the remote data
        """
        if self._raw:
            return self._raw

        path = self._existing_file_path()
        if path is not None:
            self._raw = _read(path)
            return self._raw

        if self._local_path:
            self._raw = _read(self._local_path)
            return self._raw

        if self._obj is not None:
            self._raw = serializer(self.fmt)(self._obj)
            return self._raw
        try:
            stream = self._get_stream()
            self._raw = stream.read()
            return self._raw
        finally:
            stream.close()

    def obj(self):
        if self._obj is not None:
            return self._obj
        #should be able to pass a file in later
        try:
            raw = self.raw()
            obj = deserializer(self.fmt)(raw)
            self._obj = obj
            return obj
        except Exception as e:
            logger.error("error with %s on %s raw",
                         self.data_url,
                         settings.data_rpc_url,
            )
            logger.exception(e)
            raise

    def delete(self):
        raise NotImplementedError

    def _save_stream(self):
        if self._raw:
            return len(self._raw), StringIO(self._raw)
        elif self._local_path:
            length = stat(self._local_path).st_size
            return length, open(self._local_path, "rb")
        else:
            data = serializer(self.fmt)(self._obj)
            return len(data), StringIO(data)

    def save(self, url=None, prefix=""):
        """use this function to save a NEW data object
        """
        if self.data_url:
            raise KitchenSinkError, "Dataset is already created, cannot be modified"
        if url is None:
            self.data_url = urljoin(prefix, str(uuid.uuid4()))
        else:
            self.data_url = url
        length, f = self._save_stream()
        try:
            data_type = "object" if self._obj is not None else None
            fmt = self.fmt if self._obj is not None else None
            self._put(f, data_type=data_type, fmt=fmt)
        except Exception as e:
            f.close()
            raise e

    def pipeline(self, existing=False, url=None, prefix=""):
        """pipeline this datasource across all nodes
        existing - means this object already exists in the system
        otherwise a new one will be saved and created
        """
        if existing:
            return self.pipeline_existing()
        if self.data_url and not existing:
            raise KitchenSinkError, "Dataset is already created, cannot be modified"
        if not self.data_url:
            if url is None:
                self.data_url = urljoin(prefix,
                                        str(uuid.uuid4()))
            else:
                self.data_url = url
        length, f = self._save_stream()
        try:
            host = self.rpc_url
            client = self._pipeline_existing(host, length=length)
            self._put(f)
            return client.br()
        finally:
            f.close()

    def _pipeline_existing(self, starting_host, length=None):
        c = self.client()
        active_hosts, results = c.data_info([self.data_url])
        host_info, data_info = results[self.data_url]
        if length is None:
            length = data_info['size']
        hosts = c.call('hosts', _async=False)
        hosts = [x for x in hosts if x not in host_info]
        hosts = [x for x in hosts if x != starting_host]
        hosts.append(starting_host)
        print hosts
        for idx in range(1, len(hosts)):
            c.bc('chunked_copy', self.data_url, length, hosts[idx],
                 _queue_name=hosts[idx - 1],
            )
        c.execute()
        return c

    def pipeline_existing(self):
        c = self.client()
        host = c.pick_host(self.data_url)
        client = self._pipeline_existing(host)
        client.execute()
        return client.br()

    def __repr__(self):
        if self.data_url:
            return "RemoteData(data_url='%s')" % self.data_url
        elif self._local_path:
            return "RemoteData(local_path='%s')" % self._local_path
        else:
            return "RemoteData(obj=obj)"

    def __getitem__(self, arg):
        if settings.catalog:
            print ("loading direct")
            obj = do(self.obj()[arg])
            obj.save()
            return obj
        else:
            c = Client(self.rpc_url)
            return c.async_result(c.call(self.__class__.__getitem__, self, arg))

    __getitem__.ks_memoize = True

def du(url):
    return RemoteData(data_url=url)
def dp(path):
    return RemoteData(local_path=path)
def do(obj):
    return RemoteData(obj=obj)
def dr(raw):
    return RemoteData(raw=raw)



from .serialization import json_serialization, dill_serialization, pickle_serialization

node_url = None
formats = {}
def register_serialization(name, serializer, deserializer):
    formats[name] = (serializer, deserializer)

def register_default_serialization():
    register_serialization('json',
                           json_serialization.serialize,
                           json_serialization.deserialize)
    register_serialization('dill',
                           dill_serialization.serialize,
                           dill_serialization.deserialize)
    register_serialization('pickle',
                           pickle_serialization.serialize,
                           pickle_serialization.deserialize)
register_default_serialization()

def serializer(fmt):
    return formats[fmt][0]

def deserializer(fmt):
    return formats[fmt][1]

# we're sending function call metadata as json, and
# the rest as whatever format the user wants to send
# separator splits it out
separator = '""""'
chunk_size = 2000000

#toset
catalog = None
datadir = None
host_url = None
redis_conn = None
prefix = ""
timeout = 10
is_server = None

rpc_url = None
data_rpc_url = None

def setup_server(_redis_conn, _datadir, _host_url, _catalog):
    if not _host_url.endswith("/"):
        _host_url += "/"
    global catalog
    global datadir
    global host_url
    global rpc_url
    global redis_conn
    global data_rpc_url
    global is_server

    is_server = True
    catalog = _catalog
    datadir = _datadir
    host_url = _host_url
    data_rpc_url = host_url
    rpc_url = host_url
    redis_conn = _redis_conn

def setup_client(_rpc_url):
    if not _rpc_url.endswith("/"):
        _rpc_url += "/"
    global rpc_url
    global data_rpc_url
    global is_server
    is_server = False
    rpc_url = _rpc_url
    data_rpc_url = rpc_url

def client():
    from .clients import http
    return http.Client(rpc_url)



from six.moves.urllib.parse import urlparse, parse_qs, urlencode
from logging.config import dictConfig
import logging
try:
    import gevent
except:
    gevent = None

def parse_redis_connection(url):
    split = urlparse(url)
    protocol = split.scheme
    netloc = split.netloc
    path = split.path
    db = int(parse_qs(split.query)['db'][0])
    if protocol == 'tcp':
        host, port = netloc.split(":")
        port = int(port)
        return {'protocol' : protocol,
                'host' : host,
                'port' : port,
                'db' : db}
    elif protocol == 'unix':
        #not supported yet
        return {'protocol' : protocol,
                'path' : path,
                'db' : db}

def make_query_url(url, data):
    qs = urlencode(data)
    return url + "?" + qs

def update_dictionaries(*dictionaries):
    result = {}
    for d in dictionaries:
        result.update(d)
    return result

def send_file(file_or_buffer):
    from flask import Response
    chunksize=10000
    def generator():
        with open(file_or_buffer, "rb") as f:
            while True:
                result = f.read(chunksize)
                if not result:
                    break
                else:
                    yield result
                    if gevent:
                        gevent.sleep(0)
    return Response(generator(),
                    mimetype='application/octet-stream')
def setup_loghandlers(level=None):
    logging.getLogger("requests.packages.urllib3.connectionpool").setLevel(logging.WARNING)
    if not logging._handlers:
        dictConfig({
            'version': 1,
            'disable_existing_loggers': False,

            'formatters': {
                'console': {
                    'format' : "%(created)f:%(process)d:%(name)s:%(message)s"
                },
            },

            'handlers': {
                'console': {
                    'level': 'DEBUG',
                    'class': 'logging.StreamHandler',
                    'formatter': 'console',
                },
            },

            'root': {
                'handlers': ['console'],
                'level': level or 'INFO',
            }
        })



from os.path import join, dirname, split, realpath, exists, abspath
import posixpath


def urlsplit(path, basepath, maxdepth=10):
    """splits /home/hugo/foo/bar/baz into foo, bar, baz, assuming
    /home/hugo is the basepath
    """
    if maxdepth == 0:
        return []
    if path == basepath:
        return []
    else:
        urlpath, path = posixpath.split(path)
        results = urlsplit(urlpath, basepath, maxdepth=maxdepth - 1)
        results.append(path)
        return results

def _dirsplit(path, basepath, maxdepth=10):
    if maxdepth == 0:
        return []
    if path == basepath:
        return []
    else:
        dirpath, path = split(path)
        if path == "":
            return []
        else:
            results = dirsplit(dirpath, basepath, maxdepth=maxdepth - 1)
            results.append(path)
            return results

def dirsplit(path, basepath, maxdepth=10):
    """splits /home/hugo/foo/bar/baz into foo, bar, baz, assuming
    /home/hugo is the basepath
    """
    path = realpath(path)
    basepath = realpath(basepath)
    return _dirsplit(path, basepath, maxdepth=maxdepth)

def urljoin(*paths):
    return posixpath.join(*paths)



from functools import wraps
from ..data import RemoteData
from ..data import datarpc
from ..data.routing import inspect
from .. import settings

def memoize(func):
    func.ks_memoize = True
    return func

def reconcile(obj, infos):
    if isinstance(obj, RemoteData):
        if obj._obj is not None:
            return obj.obj()
        elif obj._local_path is not None:
            return obj.local_path()
        elif infos[-1][obj.data_url][1].get('data_type', 'object') == 'object':
            obj = obj.obj()
        else:
            obj = obj.local_path()
    return obj


def remote(func):
    """This decorator makes it easy to write functions that can be run
    locally on the client or the server.
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        urls = inspect(args, kwargs)
        if settings.is_server:
            infos = datarpc.get_info_bulk(urls)
        else:
            c = settings.client()
            infos = c.data_info(urls)
        new_args = []
        for idx, arg in enumerate(args):
            arg = reconcile(arg, infos)
            new_args.append(arg)
        for k, v in kwargs.items():
            kwargs[k] = reconcile(v, infos)
        result = func(*new_args, **kwargs)
        assert isinstance(result, RemoteData)
        return result
    wrapper.ks_remote = True
    return wrapper





def inspect(func_args, func_kwargs):
    #fixme : circular import
    from . import RemoteData
    urls = set()
    all_args = list(func_args) + func_kwargs.values()
    for arg in all_args:
        if isinstance(arg, RemoteData):
            urls.add(arg.data_url)
    return urls


def route(urls, hosts, infos, threshold):
    """urls : list of data urls
    hosts: list of host urls
    infos: dict of data_urls to tuple of (host_info, data_info)
    """
    to_copy = {} #dict of host -> how much data needs to be copied
    for host in hosts:
        to_copy[host] = 0
    #M*N algo.. should be ok cause M and N are small
    for url in urls:
        host_info, data_info = infos[url]
        size = data_info['size']
        for host in hosts:
            if host_info.get(host):
                continue
            else:
                to_copy[host] += size
    to_copy = to_copy.items()
    to_copy.sort(key=lambda x : x[1])
    to_copy = [x for x in to_copy if x[1] < threshold]
    return to_copy



import pandas as pd
import cStringIO as StringIO
def serialize(obj):
    assert isinstance(obj, pd.DataFrame)
    f = StringIO.StringIO
    obj.to_csv(f)
    f.seek(0)
    return f.read()

def deserialize():
    f = StringIO.StringIO
    f.write(obj)
    f.seek(0)
    return pd.read_csv(f)



from __future__ import print_function
import time
import random
import datetime as dt
import logging

import six
import requests
import dill
from rq.job import Status

from ..serialization import (serializer, deserializer, pack_msg,
                             unpack_result, unpack_msg,
                             pack_rpc_call, unpack_results)
from ..utils import make_query_url
from ..errors import KitchenSinkError
from .. import settings

logger = logging.getLogger(__name__)

class Client(object):
    """This class contains a variety of methods for calling functions
    The convention for kwargs that are used by the
    kitchensink framework (and which do not get passed to the
    underlying function) is to preface them with a "_"

    _queue_name - route the call to a sepcific queue
    _async - execute the task asynchronously
    _data_info - previoulsy retrieved metadata about remote data sources
    _active-hosts - previously retrieved information about which
      hosts are active
    _rpc_name - which rpc namespace the task should be executed in
    _no_route_data - don't try to use data locality based routing
    _intermediate_results - whether or not to pass stdout back to the client
    _prefix - prefix used in saving remote data (used by experimental decorators)

    For ease of interactive use, we store a decent amount of state on the client itself
    self.bulk_call, will store a function call specification (func, args, kwargs)
    in self.calls
    self.execute will execute all calls in self.calls, and store the resulting
    job ids in self.jids
    self.bulk_results will retrieving the results in self.jids, and store
    the result in self.results (And return them)

    We could make it so you didn't have to use the client in such a way that
    all the state is stored on the client, but there really is no downside to this
    clients are cheap to construct, so, I think the mode of operation should just
    be to clone this one if we need a new one
    """
    def __init__(self, url, rpc_name='default',
                 queue_name="default",
                 fmt='cloudpickle'):
        # client settings
        self.url = url
        self.fmt = fmt
        self.rpc_name = rpc_name
        self.queue_name = queue_name
        self.data_threshold = 200000000 #200 megs
        self.local = False
        self.prefix = ""

        # client state, used in the bulk call pipeline
        self.calls = []
        self.jids = []
        self.results = []

    def clone(self):
        """constructs a client with the same settings as this one, but
        without job state (self.calls, self.jids, self.results)
        """
        return self.__class__(self.url, rpc_name=self.rpc_name,
                              queue_name=self.queue_name,
                              fmt=self.fmt)

    """Functions for executing many calls on the server
    """
    def bulk_call(self, func, *args, **kwargs):
        """bulk_call will store a function call on this instance
        to be executed by calling execute
        """
        if self.prefix and '_prefix' not in kwargs:
            kwargs['_prefix'] = self.prefix
        self.calls.append((func, args, kwargs))

    bc = bulk_call

    def _bulk_execute(self, calls, active_hosts, data_info):
        """execute a large number of async calls on the server
        calls - tuple of (func, args, kwargs)
        active_hosts - set of active hosts
        data_info - dict of url -> metadata about the data
        """
        data = []
        for func, args, kwargs in calls:
            rpc_name = kwargs.pop('_rpc_name', self.rpc_name)
            kwargs['_data_info'] = data_info
            kwargs['_active_hosts'] = active_hosts
            msg = self.rpc_message(func, *args, **kwargs)
            data.append(rpc_name),
            data.append(msg)
        msg = pack_msg(*data, fmt=["raw" for x in range(len(data))])
        url = self.url + "rpc/bulkcall/"
        result = requests.post(url, data=msg,
                               headers={'content-type' : 'application/octet-stream'})
        msg_format, messages = unpack_msg(result.content, override_fmt='raw')
        jids = []
        for msg in messages:
            msg_format, [metadata, data] = unpack_result(msg)
            if metadata['status'] == Status.FAILED:
                raise Exception, metadata['error']
            else:
                jids.append(metadata['job_id'])
        return jids

    def execute(self):
        """execute all calls queued up with bulk_call
        """
        if self.local:
            return self.execute_local()
        urls = set()
        from ..data.routing  import inspect, route
        for func, args, kwargs in self.calls:
            urls.update(inspect(args, kwargs))
        if urls:
            active_hosts, data_info = self.data_info(urls)
        else:
            active_hosts, data_info = None, None
        calls = self.calls
        self.calls = []
        self.jids = self._bulk_execute(calls, active_hosts, data_info)

    def bulk_async_result(self, job_ids, timeout=6000.0):
        to_query = job_ids
        raw_url = self.url + "rpc/bulkstatus/"
        results = {}
        st = time.time()
        while True:
            to_query = list(set(to_query).difference(set(results.keys())))
            print ("WAITING ON %s" % len(to_query))
            if time.time() - st > timeout:
                break
            if len(to_query) == 0:
                break

            result = requests.post(raw_url,
                                   data={'job_ids' : ",".join(to_query)},
            )
            metadata_data_pairs = unpack_results(result.content)
            for job_id, (metadata, data) in zip(to_query, metadata_data_pairs):
                for msg in metadata.get('msgs', []):
                    if msg['type'] == 'status':
                        pass
                    else:
                        print (msg['msg'])
                if metadata['status'] == Status.FAILED:
                    self.bulk_cancel(to_query)
                    raise Exception(data)
                elif metadata['status'] == Status.FINISHED:
                    results[job_id] = data
                else:
                    pass
        return [results[x] for x in job_ids]

    def bulk_cancel(self, jids=None):
        if jids is None:
            jids = self.jids
        raw_url = self.url + "rpc/bulkcancel/"
        result = requests.post(raw_url,
                               data={'job_ids' : ",".join(jids)}
        )

    def cancel_all(self):
        return self.call('cancel_all', _rpc_name='admin', _async=False)

    """functions for executing a call locally (on this machine)
    these are mostly experimental
    """
    def execute_local(self):
        self.results = []
        for c in self.calls:
            func, args, kwargs = c
            r = func(*args, **kwargs)
            self.results.append(r)

    def bulk_results_local(self):
        return self.results

    def bulk_results(self):
        if self.local:
            return self.bulk_results_local()
        try:
            return self.bulk_async_result(self.jids)
        except KeyboardInterrupt as e:
            self.bulk_cancel(self.jids)

    br = bulk_results


    def rpc_message(self, func, *args, **kwargs):
        #TODO: check for serialized function
        #TODO: handle instance methods
        async = kwargs.pop('_async', True)
        active_hosts = kwargs.pop('_active_hosts', None)
        data_info = kwargs.pop('_data_info', None)
        no_route_data = kwargs.pop('_no_route_data', False)
        intermediate_results = kwargs.pop('_intermediate_results', True)
        prefix = kwargs.pop('_prefix', "")

        func_string = None
        if isinstance(func, six.string_types):
            func_string = func
        else:
            func = func
        #pass func in to data later, when we support that kind of stuff
        if "_queue_name" in kwargs:
            queue_names = [kwargs.pop("_queue_name")]
        elif no_route_data:
            queue_names = [self.queue_name]
        else:
            #fixme circular import
            from ..data.routing  import inspect, route
            data_urls = inspect(args, kwargs)
            if data_urls:
                if active_hosts is None and data_info is None:
                    active_hosts, data_info = self.data_info(data_urls)
                queue_names = route(data_urls, active_hosts, data_info,
                                    self.data_threshold)
                #strip off size information
                queue_names = [x[0] for x in queue_names]
                logger.debug("routing to %s", queue_names)
            else:
                queue_names = [self.queue_name]
        fmt = self.fmt
        auth_string = ""

        metadata = dict(
            intermediate_results=intermediate_results,
            func_string=func_string,
            result_fmt=fmt,
            queue_names=queue_names,
            auth_string=auth_string,
            async=async,
            prefix=prefix
        )

        data = dict(func=func,
                    args=args,
                    kwargs=kwargs)
        msg = pack_rpc_call(metadata, data, fmt=self.fmt)
        return msg

    """
    functions for executing one function, or retrieving one result,
    one at a time
    """

    def call(self, func, *args, **kwargs):
        async = kwargs.get('_async', True)
        rpc_name = kwargs.pop('_rpc_name', self.rpc_name)
        msg = self.rpc_message(func, *args, **kwargs)
        url = self.url + "rpc/call/%s/" % rpc_name
        result = requests.post(url, data=msg,
                               headers={'content-type' : 'application/octet-stream'})
        msg_format, [metadata, data] = unpack_result(result.content)
        if metadata['status'] == Status.FAILED:
            raise Exception, metadata['error']
        elif metadata['status'] == Status.FINISHED:
            return data
        elif async:
            return metadata['job_id']

    def async_result(self, jobid, retries=10):
        for c in range(retries):
            url = self.url + "rpc/status/%s" % jobid
            result = requests.get(url,
                                  headers={'content-type' : 'application/octet-stream'})
            msg_format, [metadata, data] = unpack_result(result.content)
            for msg in metadata.get('msgs', []):
                if msg['type'] == 'status':
                    print (msg['status'])
                else:
                    print (msg['msg'])
            if metadata['status'] == Status.FAILED:
                raise Exception(data)
            elif metadata['status'] == Status.FINISHED:
                return data
            elif metadata['status'] == Status.STARTED:
                pass


    def cancel(self, jobid):
        """cannot currently cancel running jobs
        """
        url = self.url + "rpc/cancel/%s" % jobid
        resp = requests.get(url)
        if resp.status_code == 200:
            return
        else:
            raise KitchenSinkError("Failed to cancel %s" % jobid)

    """remote data functions
    """
    def data_info(self, urls):
        active_hosts, results = self.call('get_info_bulk', urls,
                                          _async=False, _rpc_name="data",
                                          _no_route_data=True)
        return active_hosts, results

    def _get_data(self, path, offset=None, length=None):
        url = self.url + "rpc/data/%s/" % path
        if offset is not None and length is not None:
            url = make_query_url(url, {'offset' : offset, 'length' : length})
        result = requests.get(url, stream=True)
        return result

    def _put_data(self, path, f, data_type='object', fmt="cloudpickle"):
        url = self.url + "rpc/data/%s/" % path
        result = requests.post(url,
                               data={'data_type' : data_type,
                                     'fmt' : fmt
                                 },
                               files={'data' : f}).json()
        if result.get('error'):
            raise Exception(result.get('error'))
        return result

    def pick_host(self, data_url):
        host_info, data_info = self.call('get_info', data_url,
                                         _rpc_name='data',
                                         _async=False)
        if settings.host_url and settings.host_url in host_info:
            return host_info[settings.host_url]
        host = host_info.keys()[0]
        return host

    def path_search(self, pattern):
        return self.async_result(self.call('search_path', pattern,
                                           _async=True,
                                           _rpc_name='data'))

    def reducetree(self, pattern, number=0, remove_host=None):
        matching = self.path_search(pattern)
        self.reducedata(*matching, number=number, remove_host=remove_host)

    def reducedata(self, *urls, **kwargs):
        c = self.clone()
        number = kwargs.pop('number', 0)
        remove_host = kwargs.pop('remove_host', None)
        active_hosts, infos = self.data_info(urls)
        for url in urls:
            host_info, data_info = infos[url]
            hosts = set(active_hosts).intersection(set(host_info.keys()))
            if not len(hosts) > number:
                print("%s hosts have data for %s.  Not reducing" % (len(hosts), url))
                continue
            if remove_host and remove_host in hosts:
                hosts = [x for x in hosts if x != remove_host]
                hosts.append(remove_host)
            to_keep = set(list(hosts)[:number])
            assert len(to_keep) == number
            to_delete = set(hosts).difference(to_keep)
            for host in to_delete:
                c.bulk_call('delete', url, _queue_name=host,
                            _rpc_name='data')
        c.execute()
        return c.bulk_results()

    """
    experimental data management operations - might not work
    use at your own peril
    """

    def move_data_pattern(self, pattern, from_host, to_host=None):
        c = self.clone()
        urls = self.path_search(pattern)
        info = self.data_info(urls)
        active_hosts, data_info = info

        candidates = [x for x in active_hosts if x != from_host]
        for url, (host_info, metadata) in data_info.items():
            if to_host is None:
                target = random.choice(candidates)
            else:
                target = to_host
            if from_host in host_info and len(host_info) == 1:
                c.bc('chunked_copy', url, metadata.get('size'), from_host, _queue_name=target)
        c.execute()
        c.br()

        info = self.data_info(urls)
        active_hosts, data_info = info
        for url, (host_info, metadata) in data_info.items():
            if from_host in host_info and len(host_info) >= 2:
                self.bc('delete', url, _queue_name=from_host, _rpc_name='data')
        self.execute()
        self.br()

    def md5(self, url):
        from kitchensink.data import du
        infos = self.data_info([url])
        active_hosts = infos[0]
        info, metadata = infos[-1][url]
        hosts = set(active_hosts).intersection(set(info.keys()))
        obj = du(url)
        for h in hosts:
            self.bc(md5sum, obj, _queue_name=h)
        self.execute()
        results = self.br()
        return zip(hosts, results)

    def ping(self):
        url = self.url + "rpc/ping/"
        try:
            return requests.get(url).content
        except requests.ConnectionError:
            return None

import hashlib
def md5sum(obj):
    path = obj.local_path()
    m = hashlib.md5()
    with open(path) as f:
        m.update(f.read())
    return m.hexdigest()



import copy
import time
import logging

from rq.job import Status
from rq import Queue, cancel_job
from rq import Connection

from .objs import KitchenSinkJob, KitchenSinkRedisQueue, pull_intermediate_results

log = logging.getLogger(__name__)

class TaskQueue(object):
    def __init__(self, redis_conn):
        self.conn = redis_conn
        self.queues = {}

    def get_queue(self, name):
        if not name in self.queues:
            queue = KitchenSinkRedisQueue(name, connection=self.conn)
            self.queues[name] = queue
        return self.queues[name]

    def make_job(self, func, args, kwargs, metadata={}):
        job = KitchenSinkJob.create(
            func, args, kwargs, connection=self.conn,
            result_ttl=None, status=Status.QUEUED,
            description="ksjob", depends_on=None, timeout=None)
        for k,v in metadata.items():
            job.meta[k] = v
        job.save()
        return job

    def enqueue(self, queue_names, func, args, kwargs, metadata={}):
        job = self.make_job(func, args, kwargs, metadata=metadata)
        for queue_name in queue_names:
            queue = self.get_queue(queue_name)
            queue.enqueue_job(job)
        return job._id, job.get_status()

    # def enqueue(self, queue_name, func, args, kwargs, metadata={}):
    #     queue = self.get_queue(queue_name)
    #     job = self.get_queue(queue_name).enqueue_call(func, args=args, kwargs=kwargs)
    #     for k,v in metadata.items():
    #         job.meta[k] = v
    #     job.save()
    #     return job._id, job.get_status()

    def bulkstatus(self, job_ids, timeout=5.0):
        st = time.time()
        with Connection(self.conn):
            jobs = [KitchenSinkJob(job_id) for job_id in job_ids]
        statuses = [job.get_status() for job in jobs]
        if any([x in {Status.FINISHED, Status.FAILED} for x in statuses]):
            messages = pull_intermediate_results(self.conn, job_ids, timeout=0)
        else:
            messages = pull_intermediate_results(self.conn, job_ids, timeout=timeout)
        mt = time.time()
        log.info("GOT MESSAGES %s, %s, %s" % (mt, mt - st, len(messages)))
        statuses = [job.get_status() for job in jobs]
        metadata = {}
        results = {}
        messages_by_job = {}
        for job_id, msg in messages:
            messages_by_job.setdefault(job_id, []).append(msg)
        for job_id, job, status in zip(job_ids, jobs, statuses):
            if status == Status.FINISHED or status == Status.FAILED:
                job.refresh()
                mdata = copy.copy(job.meta)
            else:
                mdata = {}
            mdata['status'] = status
            mdata['msgs'] = messages_by_job.get(job_id, [])
            metadata[job_id] = mdata
            if status == Status.FINISHED:
                results[job_id] = job.return_value
                job.delete()
            elif status == Status.FAILED:
                results[job_id] = job.exc_info
                job.delete()
            else:
                results[job_id] = None
        ed = time.time()
        log.debug("queue BULKSTATUS %s", ed - st)
        return [(metadata[job_id], results[job_id]) for job_id in job_ids]

    def status(self, job_id, timeout=None):
        """timeout is None - default to non-blocking
        """
        with Connection(self.conn):
            job = KitchenSinkJob(job_id)
        job.refresh()
        status = job.get_status()
        # if the job is finished, grab all messages and get out!
        if status == Status.FINISHED:
            messages = job.pull_intermediate_results(timeout=0)
        else:
            messages = job.pull_intermediate_results()
            #then refresh the job again
            job.refresh()
            status = job.get_status()
        metadata = copy.copy(job.meta)
        metadata['status'] = status
        metadata['msgs'] = messages

        if status == Status.FINISHED:
            retval = job.return_value
            job.delete()
            return metadata, retval

        elif status == Status.FAILED:
            retval = job.exc_info
            job.delete()
            return metadata, retval
        else:
            return metadata, None
    def cancel(self, job_id):
        with Connection(self.conn):
            cancel_job(job_id)



import logging
from os.path import exists
import traceback
import time

from flask import request, current_app, jsonify
from rq.job import Status

from .app import rpcblueprint
from ..serialization import pack_result, pack_results, unpack_msg, pack_msg
from .. import settings
from ..utils import send_file

logger = logging.getLogger(__name__)

# we assume that you set rpc.rpc to some instance of an RPC object
def make_json(jsonstring, status_code=200, headers={}):
    """like jsonify, except accepts string, so we can do our own custom
    json serialization.  should move this to continuumweb later
    """
    return current_app.response_class(response=jsonstring,
                                      status=status_code,
                                      headers=headers,
                                      mimetype='application/json')


### job endpoints
@rpcblueprint.route("/call/<rpcname>/", methods=['POST'])
def call(rpcname):
    logger.info("POST, call")
    msg = request.data
    rpc = rpcblueprint.rpcs[rpcname]
    result = rpc.call(msg)
    logger.info("POST, done call")
    return current_app.response_class(response=result,
                                      status=200,
                                      mimetype='application/octet-sream')

@rpcblueprint.route("/bulkcall/", methods=['POST'])
def bulkcall():
    logger.info("BULKPOST, call")
    msg = request.data
    st = time.time()
    msg_format, messages = unpack_msg(msg, override_fmt='raw') # strip of format
    ed = time.time()
    logger.info("unpack, msg %f", ed-st)
    calls = [(messages[i], messages[i+1]) for i in range(0, len(messages), 2)]
    results = []
    msg_index = {}
    msgs_by_rpc = {}
    all_results = {}
    for idx, (rpcname, msg) in enumerate(calls):
        msgs_by_rpc.setdefault(rpcname, []).append((idx, msg))
    for rpcname in msgs_by_rpc.keys():
        rpc = rpcblueprint.rpcs[rpcname]
        temp = msgs_by_rpc[rpcname]
        idxs = [x[0] for x in temp]
        msgs = [x[1] for x in temp]
        results = rpc.bulk_call(msgs)
        for r, idx, msg in zip(results, idxs, msgs):
            all_results[idx] = r
    all_results = [all_results[x] for x in range(len(calls))]
    result = pack_msg(*all_results, fmt=['raw' for x in range(len(all_results))])
    return current_app.response_class(response=result,
                                      status=200,
                                      mimetype='application/octet-sream')

@rpcblueprint.route("/status/<job_id>/")
def status(job_id):
    timeout = request.values.get('timeout')
    if timeout:
        timeout = float(timeout)
    metadata, value = rpcblueprint.task_queue.status(job_id, timeout=timeout)
    result = pack_result(metadata, value, fmt=metadata['result_fmt'])
    return current_app.response_class(response=result,
                                      status=200,
                                      mimetype='application/octet-stream')

@rpcblueprint.route("/cancel/<job_id>/")
def cancel(job_id):
    rpcblueprint.task_queue.cancel(job_id)
    return "success"

@rpcblueprint.route("/bulkcancel/", methods=['POST'])
def bulk_cancel():
    job_ids = request.values.get('job_ids').split(",")
    for job_id in job_ids:
        rpcblueprint.task_queue.cancel(job_id)
    return "success"

@rpcblueprint.route("/bulkstatus/", methods=['POST', 'GET'])
def bulk_status():
    logger.info("START BULKSTATUS, call")
    timeout = request.values.get('timeout', 5)
    job_ids = request.values.get('job_ids').split(",")
    if timeout:
        timeout = int(timeout)
    metadata_data_pairs = rpcblueprint.task_queue.bulkstatus(job_ids, timeout=timeout)
    # hack, with actual results, result_fmt should be present
    # otherwise for status we just use json
    fmt = [x[0].get('result_fmt', 'json') for x in metadata_data_pairs]
    result = pack_results(metadata_data_pairs, fmt=fmt)
    logger.info("DONE BULKSTATUS, call")
    return current_app.response_class(response=result,
                                      status=200,
                                      mimetype='application/octet-stream')

### Data endpoints

@rpcblueprint.route("/data/<path:path>/", methods=['GET'])
def get_data(path):
    #check auth here if we're doing auth
    offset = request.values.get('offset')
    length = request.values.get('length')
    if offset is not None and length is not None:
        local_path = settings.catalog.get_file_path(path, unfinished=True)
        offset = int(offset)
        length = int(length)
        if local_path is None or not exists(local_path):
            data = b""
        else:
            with open(local_path, "rb") as f:
                f.seek(offset)
                data = f.read(length)
        logger.info("sending %s of %s", len(data), path)
        return current_app.response_class(response=data,
                                          status=200,
                                          mimetype='application/octet-stream')
    else:
        local_path = settings.catalog.get_file_path(path)
        logger.info("sending %s", path)
        return send_file(local_path)

@rpcblueprint.route("/data/<path:path>/", methods=['POST'])
def put_data(path):
    #check auth here if we're doing auth
    logger.info("PUT")
    fstorage = request.files['data']
    data_type = request.values.get('data_type', 'object')
    fmt = request.values.get('fmt', 'cloudpickle')
    try:
        settings.catalog.write(fstorage, path, is_new=True,
                               data_type=data_type,
                               fmt=fmt)
        return jsonify(success=True)
    except Exception as e:
        exc_info = traceback.format_exc()
        return jsonify(error=exc_info)

# @rpcblueprint.route("/chunkeddata/<path:path>/", methods=['GET'])
# def get_chunked_data(path):
#     #check auth here if we're doing auth
#     offset = int(request.values['offset'])
#     length = int(request.values['offset'])
#     local_path = settings.catalog.get_file_path(path)
#     with open(local_path, "rb") as f:
#         f.seek(offset)
#         data = f.read(local_path, settings.chunk_size)
#     return current_app.response_class(response=data,
#                                       status=200,
#                                       mimetype='application/octet-sream')

@rpcblueprint.route("/ping/", methods=['GET', 'POST'])
def ping():
    return 'pong'



import logging
from argparse import ArgumentParser
from rq import Connection
import redis

from kitchensink import settings
from kitchensink.utils import parse_redis_connection
from kitchensink.taskqueue.objs import KitchenSinkRedisQueue, KitchenSinkWorker
from kitchensink.data import Catalog
comments = \
"""
kitchen sink RPC Server

This command is used to start a worker process for the RPC Server.
"""
def parser():
    p = ArgumentParser(comments)
    p.add_argument('--redis-connection',
                   help="redis connection information, <ip>:<port>:db",
                   default="localhost:6379:0")
    p.add_argument('--node-url', help='url of node', default='http://localhost:6324/')
    p.add_argument('--queue',
                   help='queue to operate on',
                   action='append')
    p.add_argument('--reload',
                   help='reload when code has changed',
                   default=False,
                   action='store_true'
    )
    p.add_argument('--datadir',
                   help='data directory',
    )
    return p

def run_args(args):
    run(args.redis_connection, args.node_url, args.queue, args.datadir)

def run(redis_connection, node_url, queue, datadir):
    settings.node_url = node_url
    redis_connection_obj = parse_redis_connection(redis_connection)
    r = redis.StrictRedis(host=redis_connection_obj['host'],
                          port=redis_connection_obj['port'],
                          db=redis_connection_obj['db'])
    settings.setup_server(r, datadir, node_url,
                          Catalog(r, datadir, node_url))
    if queue is None:
        queue = ['default']
    with Connection(r):
        queues = []
        node_queue = KitchenSinkRedisQueue(node_url)
        queues.append(node_queue)
        for q in queue:
            queues.append(KitchenSinkRedisQueue(q))
            w = KitchenSinkWorker(queues, default_result_ttl=86400)
    w.work(burst=False)

def main():
    p = parser()
    args = p.parse_args()
    def helper():
        run_args(args)
    if args.reload:
        run_with_reloader(helper)
    else:
        helper()

if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    logging.getLogger("requests.packages.urllib3.connectionpool").setLevel(logging.WARNING)
    #logging.getLogger('rq.worker').setLevel(logging.WARNING)
    main()



import logging
import sys

from ..rpc import RPC
from .. import settings
logger = logging.getLogger(__name__)

def cancel_all():
    keys = settings.redis_conn.keys("rq:job*")
    if keys:
        settings.redis_conn.delete(*keys)

def make_rpc():
    rpc = RPC()
    rpc.register_function(cancel_all)
    return rpc



from ..rpc import RPC
import time
def dummy_mult(x, y):
    return x * y

def dummy_add(x, y):
    return x + y

def dummy_sleeper_add(x, y):
    print ('sleep 1')
    time.sleep(1)
    print ('sleep 2')
    time.sleep(1)
    print ('sleep 3')
    time.sleep(1)
    return x + y

def make_rpc():
    rpc = RPC()
    rpc.register_function(dummy_mult)
    rpc.register_function(dummy_add)
    rpc.register_function(dummy_sleeper_add)
    return rpc



import pandas as pd
import numpy as np
import cPickle as pickle
import json

def encode_dataframe(obj):
    pass

def encode_numpy(obj):
    pass

def encode_dtype(obj):
    # punting on properly encoding dtypes
    # in a transparent manner for now
    val = pickle.dumps(obj)
    return {'type' : 'dtype',
            'val' : val}
    
def decode_dtype(obj):
    return pickle.loads(str(obj['val']))

class KitchenSinkEncoder(json.JSONEncoder):
    def default(self, obj):
        ## array types
        if isinstance(obj, np.dtype):
            return encode_dtype(obj)
            
def object_hook(obj):
    if 'type' in obj:
        if obj['type'] == 'dtype':
            return decode_dtype(obj)
    return obj

def serialize(obj, encoder=KitchenSinkEncoder, **kwargs):
    return json.dumps(obj, cls=encoder, **kwargs)

def deserialize(obj, **kwargs):
    return json.loads(obj, object_hook=object_hook, **kwargs)



import atexit
import time
from threading import Thread
try:
    import gevent
except:
    gevent = None
import redis
import redis.connection
from rq import Queue, Connection

from .app import app, rpcblueprint
from . import views
from ..taskqueue import TaskQueue
from .. import settings
from ..data import Catalog

def get_queue(name):
    if not name in rpcblueprint.queues:
        with Connection(rpcblueprint.r):
            queue = Queue(name)
            rpcblueprint.queues[name] = queue
    return rpcblueprint.queues[name]

def make_app(redis_connection_obj, port, host_url, datadir):
    app.register_blueprint(rpcblueprint, url_prefix="/rpc")
    app.port = port
    if gevent:
        redis.connection.socket = gevent.socket
    rpcblueprint.r = redis.StrictRedis(host=redis_connection_obj['host'],
                                       port=redis_connection_obj['port'],
                                       db=redis_connection_obj['db'])
    rpcblueprint.task_queue = TaskQueue(rpcblueprint.r)
    settings.setup_server(rpcblueprint.r, datadir, host_url,
                          Catalog(rpcblueprint.r, datadir, host_url))
    rpcblueprint.heartbeat_thread = HeartbeatThread()
    return app

def register_rpc(rpc, name='default'):
    rpcblueprint.rpcs[name] = rpc
    rpc.setup_queue(rpcblueprint.task_queue)

def close():
    rpcblueprint.heartbeat_thread.kill = True
    rpcblueprint.heartbeat_thread.join()

def run(gevent=False):
    app.debug = True
    rpcblueprint.heartbeat_thread.start()
    atexit.register(close)
    if gevent:
         from gevent.pywsgi import WSGIServer
         http_server = WSGIServer(("0.0.0.0", app.port), app)
         http_server.serve_forever()
    else:
        app.run(host='0.0.0.0', port=app.port, use_reloader=False)
    close()

class HeartbeatThread(Thread):
    def run(self):
        self.kill = False
        if settings.prefix:
            self.host_key = settings.prefix + ":" + "hosts"
            self.hostinfo_key = settings.prefix + ":" + "hostinfo:%s" %settings.host_url
        else:
            self.host_key = "hosts"
            self.hostinfo_key = "hostinfo:%s" %settings.host_url
        def loop():
            settings.redis_conn.sadd(self.host_key, self.hostinfo_key)
            settings.redis_conn.setex(self.hostinfo_key,
                                      settings.timeout,
                                      settings.host_url)
        def remove():
            settings.redis_conn.srem(self.host_key, self.hostinfo_key)
            settings.redis_conn.delete(self.hostinfo_key, settings.host_url)
        loop()
        while True:
            if self.kill:
                break
            else:
                loop()
                time.sleep(1)
        remove()



import logging
import sys

from ..rpc import RPC
from .. import settings
from . import hosts

logger = logging.getLogger(__name__)
def get_info(path):
    #TODO: check against availble servers
    #TODO: pipeline redis operations
    host_info, data_info = settings.catalog.get_info(path)
    return host_info, data_info

def get_info_bulk(urls):
    results = {}
    active_hosts = hosts()
    for u in urls:
        host_info, data_info = get_info(u)
        for host in host_info.keys():
            if host not in active_hosts:
                host_info.pop(host)
        results[u] = host_info, data_info
    return active_hosts, results

def search_path(path_pattern):
    return settings.catalog.search(path_pattern)

def chunked_copy(url, length, host):
    #print >> sys.stderr ,"chunked copy of %s from %s to %s" % (url, host, settings.host_url)
    logger.info("chunked copy of %s from %s to %s" % (url, host, settings.host_url))
    iterator = settings.catalog.get_chunked_iterator(url, length, host)
    settings.catalog.write_chunked(iterator, url, is_new=False)

def delete(url):
    settings.catalog.delete(url)

def make_rpc():
    rpc = RPC(allow_arbitrary=False)
    rpc.register_function(get_info)
    rpc.register_function(get_info_bulk)
    rpc.register_function(search_path)
    rpc.register_function(hosts)
    rpc.register_function(chunked_copy)
    rpc.register_function(delete)
    return rpc



import traceback
import sys
import cStringIO
import threading
import logging
import time
import hashlib

from rq.job import Status
from rq import Queue, Connection
from six import string_types

from ..serialization import (json_serialization,
                             dill_serialization,
                             pickle_serialization,
                             pack_result,
                             unpack_rpc_metadata,
                             unpack_msg_format,
                             unpack_rpc_call,
                             append_rpc_data
)

from ..taskqueue.objs import current_job_id, KitchenSinkJob
from ..errors import UnauthorizedAccess, UnknownFunction, WrappedError, KitchenSinkError
from ..settings import serializer, deserializer
from .. import settings
from ..data import du, do

logger = logging.getLogger(__name__)
"""
Aspects:
- async or not
- serialized/deserialized
- currently, the serialization format stops and ends with the head node
- however in the future, serialization will be done on the worker nodes

NOTE: for security reasons, we should probably refactor this to
unpack the message on the work queue, since dill/pickle allow for arbitrary code
execution
"""
class RPC(object):
    def __init__(self,
                 task_queue=None,
                 allow_arbitrary=False,
                 default_serialization_config=True,
                 auth=None):
        """
        Args
            allow_arbitrary : allow calls of arbitrary
                functions that are not registered
            default_serialization_config : use default
                serialization configuration
        """
        self.allow_arbitrary = allow_arbitrary
        self.functions = {}
        self.task_queue = task_queue
        self.auth = auth
        self.queues = {}

    def setup_queue(self, queue):
        self.task_queue = queue

    def bulk_call(self, msgs):
        """TODO: refactor this so it's less complicated.
        what's going on is we're going through call1, call2, call3
        and then routing call1, call2, call3 to the first priority host
        and then call1, call2, call3 to the second priority host, etc..
        """
        async_jobs = {}
        async_job_metadata = {}
        async_queues = {}
        results = {}
        for idx, msg in enumerate(msgs):
            metadata = unpack_rpc_metadata(msg)
            result_fmt = metadata.get('result_fmt', 'cloudpickle')
            queue_names = metadata.get('queue_names', ['default'])
            async = metadata.get('async', True)
            auth = metadata.get('auth', '')
            if auth and self.auth:
                if not self.auth(auth):
                    raise UnauthorizedException
            try:
                if async:
                    fmt = metadata['result_fmt']
                    func_string = metadata.get('func_string')
                    if func_string is not None:
                        func = self.resolve_function(func_string)
                        if func is None:
                            raise KitchenSinkError("Unknownn function %s" % func_string)
                        msg = append_rpc_data(msg, {'func' : func}, fmt='cloudpickle')
                    args = [msg]
                    kwargs = {'intermediate_results' : metadata['intermediate_results']}
                    job = self.task_queue.make_job(execute_msg,
                                             args, kwargs,
                                             metadata=metadata)
                    async_jobs[idx] = job
                    async_job_metadata[idx] = {'result_fmt' : fmt, 'job_id' : job.id}
                    async_queues[idx] = queue_names
                else:
                    metadata, result =  self.call_instant(msg, metadata)
                    results[idx] = pack_result(metadata, result, fmt=result_fmt)
            except Exception as e:
                exc_info = traceback.format_exc()
                metadata = {'result_fmt' : result_fmt,
                            'status' : Status.FAILED,
                            'error' : exc_info}
                result = None
                results[idx] = pack_result(metadata, result, fmt=result_fmt)
        while async_queues:
            for idx in range(len(msgs)):
                queues = async_queues.get(idx)
                if not queues:
                    continue
                job = async_jobs.get(idx, None)
                queue = queues.pop(0)
                self.task_queue.get_queue(queue).enqueue_job(job)
                if not queues:
                    del async_queues[idx]
        for k,v in async_jobs.iteritems():
            metadata = async_job_metadata[k]
            metadata['status'] = v.get_status()
            results[k] = pack_result(metadata, None, fmt=metadata['result_fmt'])
        results = [results[x] for x in range(len(msgs))]
        return results

    def call(self, msg):
        logger.debug("CALL")
        metadata = unpack_rpc_metadata(msg)
        result_fmt = metadata.get('result_fmt', 'cloudpickle')
        queue_names = metadata.get('queue_names', ['default'])
        async = metadata.get('async', True)
        auth = metadata.get('auth', '')
        if auth and self.auth:
            if not self.auth(auth):
                raise UnauthorizedException
        try:
            if async:
                metadata, result = self.call_async(msg, metadata, queue_names)
            else:
                metadata, result =  self.call_instant(msg, metadata)
        except Exception as e:
            exc_info = traceback.format_exc()
            metadata = {'result_fmt' : result_fmt,
                        'status' : Status.FAILED,
                        'error' : exc_info}
            result = None
        return pack_result(metadata, result, fmt=result_fmt)

    def call_instant(self, msg, metadata):
        func_string = metadata.get('func_string')
        fmt = metadata['result_fmt']
        if func_string is not None:
            func = self.resolve_function(func_string)
            msg = append_rpc_data(msg, {'func' : func}, fmt='cloudpickle')
        result = execute_msg(msg)
        metadata = {'result_fmt' : fmt, 'status' : Status.FINISHED}
        return metadata, result

    def call_async(self, msg, metadata, queue_names):
        ## at some point, we might pass strings
        ## directly to the backend task queue, but for now
        ## we parse them into python objects, and then
        ## re-serialize them via python-rq(which uses pickle)
        fmt = metadata['result_fmt']
        func_string = metadata.get('func_string')
        if func_string is not None:
            func = self.resolve_function(func_string)
            if func is None:
                raise KitchenSinkError("Unknownn function %s" % func_string)
            msg = append_rpc_data(msg, {'func' : func}, fmt='cloudpickle')
        job_id, status = self.task_queue.enqueue(
            queue_names,
            execute_msg,
            [msg],
            {'intermediate_results' : metadata['intermediate_results']},
            metadata=metadata
        )
        metadata = {'result_fmt' : fmt,
                    'job_id' : job_id,
                    'status' : status
        }
        return metadata, None

    def resolve_function(self, func_string):
        """turn a func_string into a function
        """
        if func_string in self.functions:
            return self.functions[func_string]
        if self.allow_arbitrary:
            return self.get_arbitrary_function(func_sring)

    def get_arbitrary_function(self, func_string):
        raise NotImplementedError

    def register_function(self, func, name=None):
        if name is None:
            name = func.__name__
        self.functions[name] = func


import tempfile
class OutputThread(threading.Thread):
    interval = 0.001
    def run(self):
        with open(self.filename, "r") as toread:
            self.buf = ""
            while not self.kill:
                self.output(toread)
            self.output(toread)
    def output(self, toread):
        read = toread.read()
        if not read:
            time.sleep(self.interval)
        self.buf += read
        msgs = self.buf.split("\n")[:-1]
        for msg in msgs:
            self.job.push_stdout(msg)
        self.buf = self.buf.rsplit("\n", 1)[-1]


def _execute_msg(msg):
    logger.debug("EXECUTING")
    st = time.time()
    msg_format, metadata, data = unpack_rpc_call(msg)
    ed = time.time()
    logger.debug("UNPACKED %s", (ed - st))
    func = data['func']
    memoize_url = None

    ## some code in place for memoization decorator(experimental)
    if hasattr(func, "ks_memoize") and func.ks_memoize and settings.catalog:
        m = hashlib.md5()
        m.update(msg)
        key = m.hexdigest()
        memoize_url = "memoize/%s" % key
        hosts_info, data_info = settings.catalog.get_info(memoize_url)
        if len(hosts_info) > 0:
            logger.debug("retrieving memoized")
            return du(memoize_url).obj()

    args = data.get('args', [])
    kwargs = data.get('kwargs', {})
    result = func(*args, **kwargs)

    ## some code in place for @remote decorator(experimental)
    if hasattr(func, 'ks_remote') and func.ks_remote:
        logger.debug("METADATA %s", metadata)
        result.save(prefix=metadata.get('prefix', ''))

    ## some code in place for memoization decorator(experimental)
    if memoize_url:
        logger.debug("saving memoized")
        do(result).save(url=memoize_url)
    logger.debug("DONE EXECUTING")
    return result

def patch_loggers(output):
    patched = []
    to_patch = logging.Logger.manager.loggerDict.values()
    to_patch.append(logging.Logger.manager.root)
    for logger in to_patch:
        if not hasattr(logger, 'handlers'):
            continue
        for handler in logger.handlers:
            if isinstance(handler , logging.StreamHandler):
                patched.append((handler, handler.stream))
                handler.stream = output
    return patched

def unpatch_loggers(patched):
    for handler, stream in patched:
        handler.stream = stream

def execute_msg(msg, intermediate_results=False):

    if not current_job_id() or not intermediate_results:
        logger.debug("**jobid %s", current_job_id())
        return _execute_msg(msg)
    logger.debug("intermediate_results %s", intermediate_results)
    output = tempfile.NamedTemporaryFile(prefix="ks-").name
    output_thread = OutputThread()
    output_thread.filename = output
    output_thread.job = KitchenSinkJob(current_job_id(),
                                       connection=settings.redis_conn)
    output_thread.kill = False
    old_stdout = sys.stdout
    old_stderr = sys.stderr
    with open(output, "w+", 0) as towrite:
        sys.stdout = towrite
        sys.stderr = towrite
        patched = patch_loggers(towrite)
        try:
            output_thread.start()
            result = _execute_msg(msg)
            return result
        finally:
            sys.stdout = old_stdout
            sys.stderr = old_stderr
            unpatch_loggers(patched)
            output_thread.kill = True
            output_thread.join()

## fmt is the format that the consumer wants, and is sending args in
## for now we use our own internal serialization (probably pickle or dill)
## we can sort out the internal serialization later
## same goes for storage of server side data into files



"""
This class is defined to override standard pickle functionality

The goals of it follow:
-Serialize lambdas and nested functions to compiled byte code
-Deal with main module correctly
-Deal with other non-serializable objects

It does not include an unpickler, as standard python unpickling suffices.

This module was extracted from the `cloud` package, developed by `PiCloud, Inc.
<http://www.picloud.com>`_.

Copyright (c) 2012, Regents of the University of California.
Copyright (c) 2009 `PiCloud, Inc. <http://www.picloud.com>`_.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of California, Berkeley nor the
      names of its contributors may be used to endorse or promote
      products derived from this software without specific prior written
      permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""


import operator
import os
import pickle
import struct
import sys
import types
from functools import partial
import itertools
from copy_reg import _extension_registry, _inverted_registry, _extension_cache
import new
import dis
import traceback

#relevant opcodes
STORE_GLOBAL = chr(dis.opname.index('STORE_GLOBAL'))
DELETE_GLOBAL = chr(dis.opname.index('DELETE_GLOBAL'))
LOAD_GLOBAL = chr(dis.opname.index('LOAD_GLOBAL'))
GLOBAL_OPS = [STORE_GLOBAL, DELETE_GLOBAL, LOAD_GLOBAL]

HAVE_ARGUMENT = chr(dis.HAVE_ARGUMENT)
EXTENDED_ARG = chr(dis.EXTENDED_ARG)

import logging
cloudLog = logging.getLogger("Cloud.Transport")

try:
    import ctypes
except (MemoryError, ImportError):
    logging.warning('Exception raised on importing ctypes. Likely python bug.. some functionality will be disabled', exc_info = True)
    ctypes = None
    PyObject_HEAD = None
else:

    # for reading internal structures
    PyObject_HEAD = [
        ('ob_refcnt', ctypes.c_size_t),
        ('ob_type', ctypes.c_void_p),
    ]


try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

# These helper functions were copied from PiCloud's util module.
def islambda(func):
    return getattr(func,'func_name') == '<lambda>'

def xrange_params(xrangeobj):
    """Returns a 3 element tuple describing the xrange start, step, and len
    respectively

    Note: Only guarentees that elements of xrange are the same. parameters may
    be different.
    e.g. xrange(1,1) is interpretted as xrange(0,0); both behave the same
    though w/ iteration
    """

    xrange_len = len(xrangeobj)
    if not xrange_len: #empty
        return (0,1,0)
    start = xrangeobj[0]
    if xrange_len == 1: #one element
        return start, 1, 1
    return (start, xrangeobj[1] - xrangeobj[0], xrange_len)

#debug variables intended for developer use:
printSerialization = False
printMemoization = False

useForcedImports = True #Should I use forced imports for tracking?



class CloudPickler(pickle.Pickler):

    dispatch = pickle.Pickler.dispatch.copy()
    savedForceImports = False
    savedDjangoEnv = False #hack tro transport django environment

    def __init__(self, file, protocol=None, min_size_to_save= 0):
        pickle.Pickler.__init__(self,file,protocol)
        self.modules = set() #set of modules needed to depickle
        self.globals_ref = {}  # map ids to dictionary. used to ensure that functions can share global env

    def dump(self, obj):
        # note: not thread safe
        # minimal side-effects, so not fixing
        recurse_limit = 3000
        base_recurse = sys.getrecursionlimit()
        if base_recurse < recurse_limit:
            sys.setrecursionlimit(recurse_limit)
        self.inject_addons()
        try:
            return pickle.Pickler.dump(self, obj)
        except RuntimeError, e:
            if 'recursion' in e.args[0]:
                msg = """Could not pickle object as excessively deep recursion required.
                Try _fast_serialization=2 or contact PiCloud support"""
                raise pickle.PicklingError(msg)
        finally:
            new_recurse = sys.getrecursionlimit()
            if new_recurse == recurse_limit:
                sys.setrecursionlimit(base_recurse)

    def save_buffer(self, obj):
        """Fallback to save_string"""
        pickle.Pickler.save_string(self,str(obj))
    dispatch[buffer] = save_buffer

    #block broken objects
    def save_unsupported(self, obj, pack=None):
        raise pickle.PicklingError("Cannot pickle objects of type %s" % type(obj))
    dispatch[types.GeneratorType] = save_unsupported

    #python2.6+ supports slice pickling. some py2.5 extensions might as well.  We just test it
    try:
        slice(0,1).__reduce__()
    except TypeError: #can't pickle -
        dispatch[slice] = save_unsupported

    #itertools objects do not pickle!
    for v in itertools.__dict__.values():
        if type(v) is type:
            dispatch[v] = save_unsupported


    def save_dict(self, obj):
        """hack fix
        If the dict is a global, deal with it in a special way
        """
        #print 'saving', obj
        if obj is __builtins__:
            self.save_reduce(_get_module_builtins, (), obj=obj)
        else:
            pickle.Pickler.save_dict(self, obj)
    dispatch[pickle.DictionaryType] = save_dict


    def save_module(self, obj, pack=struct.pack):
        """
        Save a module as an import
        """
        #print 'try save import', obj.__name__
        self.modules.add(obj)
        self.save_reduce(subimport,(obj.__name__,), obj=obj)
    dispatch[types.ModuleType] = save_module    #new type

    def save_codeobject(self, obj, pack=struct.pack):
        """
        Save a code object
        """
        #print 'try to save codeobj: ', obj
        args = (
            obj.co_argcount, obj.co_nlocals, obj.co_stacksize, obj.co_flags, obj.co_code,
            obj.co_consts, obj.co_names, obj.co_varnames, obj.co_filename, obj.co_name,
            obj.co_firstlineno, obj.co_lnotab, obj.co_freevars, obj.co_cellvars
        )
        self.save_reduce(types.CodeType, args, obj=obj)
    dispatch[types.CodeType] = save_codeobject    #new type

    def save_function(self, obj, name=None, pack=struct.pack):
        """ Registered with the dispatch to handle all function types.

        Determines what kind of function obj is (e.g. lambda, defined at
        interactive prompt, etc) and handles the pickling appropriately.
        """
        write = self.write

        name = obj.__name__
        modname = pickle.whichmodule(obj, name)
        #print 'which gives %s %s %s' % (modname, obj, name)
        try:
            themodule = sys.modules[modname]
        except KeyError: # eval'd items such as namedtuple give invalid items for their function __module__
            modname = '__main__'

        if modname == '__main__':
            themodule = None

        if themodule:
            self.modules.add(themodule)

        if not self.savedDjangoEnv:
            #hack for django - if we detect the settings module, we transport it
            django_settings = os.environ.get('DJANGO_SETTINGS_MODULE', '')
            if django_settings:
                django_mod = sys.modules.get(django_settings)
                if django_mod:
                    cloudLog.debug('Transporting django settings %s during save of %s', django_mod, name)
                    self.savedDjangoEnv = True
                    self.modules.add(django_mod)
                    write(pickle.MARK)
                    self.save_reduce(django_settings_load, (django_mod.__name__,), obj=django_mod)
                    write(pickle.POP_MARK)


        # if func is lambda, def'ed at prompt, is in main, or is nested, then
        # we'll pickle the actual function object rather than simply saving a
        # reference (as is done in default pickler), via save_function_tuple.
        if islambda(obj) or obj.func_code.co_filename == '<stdin>' or themodule is None:
            #Force server to import modules that have been imported in main
            modList = None
            if themodule is None and not self.savedForceImports:
                mainmod = sys.modules['__main__']
                if useForcedImports and hasattr(mainmod,'___pyc_forcedImports__'):
                    modList = list(mainmod.___pyc_forcedImports__)
                self.savedForceImports = True
            self.save_function_tuple(obj, modList)
            return
        else:   # func is nested
            klass = getattr(themodule, name, None)
            if klass is None or klass is not obj:
                self.save_function_tuple(obj, [themodule])
                return

        if obj.__dict__:
            # essentially save_reduce, but workaround needed to avoid recursion
            self.save(_restore_attr)
            write(pickle.MARK + pickle.GLOBAL + modname + '\n' + name + '\n')
            self.memoize(obj)
            self.save(obj.__dict__)
            write(pickle.TUPLE + pickle.REDUCE)
        else:
            write(pickle.GLOBAL + modname + '\n' + name + '\n')
            self.memoize(obj)
    dispatch[types.FunctionType] = save_function

    def save_function_tuple(self, func, forced_imports):
        """  Pickles an actual func object.

        A func comprises: code, globals, defaults, closure, and dict.  We
        extract and save these, injecting reducing functions at certain points
        to recreate the func object.  Keep in mind that some of these pieces
        can contain a ref to the func itself.  Thus, a naive save on these
        pieces could trigger an infinite loop of save's.  To get around that,
        we first create a skeleton func object using just the code (this is
        safe, since this won't contain a ref to the func), and memoize it as
        soon as it's created.  The other stuff can then be filled in later.
        """
        save = self.save
        write = self.write

        # save the modules (if any)
        if forced_imports:
            write(pickle.MARK)
            save(_modules_to_main)
            #print 'forced imports are', forced_imports

            forced_names = map(lambda m: m.__name__, forced_imports)
            save((forced_names,))

            #save((forced_imports,))
            write(pickle.REDUCE)
            write(pickle.POP_MARK)

        code, f_globals, defaults, closure, dct, base_globals = self.extract_func_data(func)

        save(_fill_function)  # skeleton function updater
        write(pickle.MARK)    # beginning of tuple that _fill_function expects

        # create a skeleton function object and memoize it
        save(_make_skel_func)
        save((code, len(closure), base_globals))
        write(pickle.REDUCE)
        self.memoize(func)

        # save the rest of the func data needed by _fill_function
        save(f_globals)
        save(defaults)
        save(closure)
        save(dct)
        write(pickle.TUPLE)
        write(pickle.REDUCE)  # applies _fill_function on the tuple

    @staticmethod
    def extract_code_globals(co):
        """
        Find all globals names read or written to by codeblock co
        """
        code = co.co_code
        names = co.co_names
        out_names = set()

        n = len(code)
        i = 0
        extended_arg = 0
        while i < n:
            op = code[i]

            i = i+1
            if op >= HAVE_ARGUMENT:
                oparg = ord(code[i]) + ord(code[i+1])*256 + extended_arg
                extended_arg = 0
                i = i+2
                if op == EXTENDED_ARG:
                    extended_arg = oparg*65536L
                if op in GLOBAL_OPS:
                    out_names.add(names[oparg])
        #print 'extracted', out_names, ' from ', names
        return out_names

    def extract_func_data(self, func):
        """
        Turn the function into a tuple of data necessary to recreate it:
            code, globals, defaults, closure, dict
        """
        code = func.func_code

        # extract all global ref's
        func_global_refs = CloudPickler.extract_code_globals(code)
        if code.co_consts:   # see if nested function have any global refs
            for const in code.co_consts:
                if type(const) is types.CodeType and const.co_names:
                    func_global_refs = func_global_refs.union( CloudPickler.extract_code_globals(const))
        # process all variables referenced by global environment
        f_globals = {}
        for var in func_global_refs:
            #Some names, such as class functions are not global - we don't need them
            if func.func_globals.has_key(var):
                f_globals[var] = func.func_globals[var]

        # defaults requires no processing
        defaults = func.func_defaults

        def get_contents(cell):
            try:
                return cell.cell_contents
            except ValueError, e: #cell is empty error on not yet assigned
                raise pickle.PicklingError('Function to be pickled has free variables that are referenced before assignment in enclosing scope')


        # process closure
        if func.func_closure:
            closure = map(get_contents, func.func_closure)
        else:
            closure = []

        # save the dict
        dct = func.func_dict

        if printSerialization:
            outvars = ['code: ' + str(code) ]
            outvars.append('globals: ' + str(f_globals))
            outvars.append('defaults: ' + str(defaults))
            outvars.append('closure: ' + str(closure))
            print 'function ', func, 'is extracted to: ', ', '.join(outvars)

        base_globals = self.globals_ref.get(id(func.func_globals), {})
        self.globals_ref[id(func.func_globals)] = base_globals

        return (code, f_globals, defaults, closure, dct, base_globals)

    def save_global(self, obj, name=None, pack=struct.pack):
        write = self.write
        memo = self.memo

        if name is None:
            name = obj.__name__

        modname = getattr(obj, "__module__", None)
        if modname is None:
            modname = pickle.whichmodule(obj, name)

        try:
            __import__(modname)
            themodule = sys.modules[modname]
        except (ImportError, KeyError, AttributeError):  #should never occur
            raise pickle.PicklingError(
                "Can't pickle %r: Module %s cannot be found" %
                (obj, modname))

        if modname == '__main__':
            themodule = None

        if themodule:
            self.modules.add(themodule)

        sendRef = True
        typ = type(obj)
        #print 'saving', obj, typ
        try:
            try: #Deal with case when getattribute fails with exceptions
                klass = getattr(themodule, name)
            except (AttributeError):
                if modname == '__builtin__':  #new.* are misrepeported
                    modname = 'new'
                    __import__(modname)
                    themodule = sys.modules[modname]
                    try:
                        klass = getattr(themodule, name)
                    except AttributeError, a:
                        #print themodule, name, obj, type(obj)
                        raise pickle.PicklingError("Can't pickle builtin %s" % obj)
                else:
                    raise

        except (ImportError, KeyError, AttributeError):
            if typ == types.TypeType or typ == types.ClassType:
                sendRef = False
            else: #we can't deal with this
                raise
        else:
            if klass is not obj and (typ == types.TypeType or typ == types.ClassType):
                sendRef = False
        if not sendRef:
            #note: Third party types might crash this - add better checks!
            d = dict(obj.__dict__) #copy dict proxy to a dict
            if not isinstance(d.get('__dict__', None), property): # don't extract dict that are properties
                d.pop('__dict__',None)
            d.pop('__weakref__',None)

            # hack as __new__ is stored differently in the __dict__
            new_override = d.get('__new__', None)
            if new_override:
                d['__new__'] = obj.__new__

            self.save_reduce(type(obj),(obj.__name__,obj.__bases__,
                                   d),obj=obj)
            #print 'internal reduce dask %s %s'  % (obj, d)
            return

        if self.proto >= 2:
            code = _extension_registry.get((modname, name))
            if code:
                assert code > 0
                if code <= 0xff:
                    write(pickle.EXT1 + chr(code))
                elif code <= 0xffff:
                    write("%c%c%c" % (pickle.EXT2, code&0xff, code>>8))
                else:
                    write(pickle.EXT4 + pack("<i", code))
                return

        write(pickle.GLOBAL + modname + '\n' + name + '\n')
        self.memoize(obj)
    dispatch[types.ClassType] = save_global
    dispatch[types.BuiltinFunctionType] = save_global
    dispatch[types.TypeType] = save_global

    def save_instancemethod(self, obj):
        #Memoization rarely is ever useful due to python bounding
        self.save_reduce(types.MethodType, (obj.im_func, obj.im_self,obj.im_class), obj=obj)
    dispatch[types.MethodType] = save_instancemethod

    def save_inst_logic(self, obj):
        """Inner logic to save instance. Based off pickle.save_inst
        Supports __transient__"""
        cls = obj.__class__

        memo  = self.memo
        write = self.write
        save  = self.save

        if hasattr(obj, '__getinitargs__'):
            args = obj.__getinitargs__()
            len(args) # XXX Assert it's a sequence
            pickle._keep_alive(args, memo)
        else:
            args = ()

        write(pickle.MARK)

        if self.bin:
            save(cls)
            for arg in args:
                save(arg)
            write(pickle.OBJ)
        else:
            for arg in args:
                save(arg)
            write(pickle.INST + cls.__module__ + '\n' + cls.__name__ + '\n')

        self.memoize(obj)

        try:
            getstate = obj.__getstate__
        except AttributeError:
            stuff = obj.__dict__
            #remove items if transient
            if hasattr(obj, '__transient__'):
                transient = obj.__transient__
                stuff = stuff.copy()
                for k in list(stuff.keys()):
                    if k in transient:
                        del stuff[k]
        else:
            stuff = getstate()
            pickle._keep_alive(stuff, memo)
        save(stuff)
        write(pickle.BUILD)


    def save_inst(self, obj):
        # Hack to detect PIL Image instances without importing Imaging
        # PIL can be loaded with multiple names, so we don't check sys.modules for it
        if hasattr(obj,'im') and hasattr(obj,'palette') and 'Image' in obj.__module__:
            self.save_image(obj)
        else:
            self.save_inst_logic(obj)
    dispatch[types.InstanceType] = save_inst

    def save_property(self, obj):
        # properties not correctly saved in python
        self.save_reduce(property, (obj.fget, obj.fset, obj.fdel, obj.__doc__), obj=obj)
    dispatch[property] = save_property

    def save_itemgetter(self, obj):
        """itemgetter serializer (needed for namedtuple support)
        a bit of a pain as we need to read ctypes internals"""
        class ItemGetterType(ctypes.Structure):
            _fields_ = PyObject_HEAD + [
                ('nitems', ctypes.c_size_t),
                ('item', ctypes.py_object)
            ]


        itemgetter_obj = ctypes.cast(ctypes.c_void_p(id(obj)), ctypes.POINTER(ItemGetterType)).contents
        return self.save_reduce(operator.itemgetter, (itemgetter_obj.item,))

    if PyObject_HEAD:
        dispatch[operator.itemgetter] = save_itemgetter



    def save_reduce(self, func, args, state=None,
                    listitems=None, dictitems=None, obj=None):
        """Modified to support __transient__ on new objects
        Change only affects protocol level 2 (which is always used by PiCloud"""
        # Assert that args is a tuple or None
        if not isinstance(args, types.TupleType):
            raise pickle.PicklingError("args from reduce() should be a tuple")

        # Assert that func is callable
        if not hasattr(func, '__call__'):
            raise pickle.PicklingError("func from reduce should be callable")

        save = self.save
        write = self.write

        # Protocol 2 special case: if func's name is __newobj__, use NEWOBJ
        if self.proto >= 2 and getattr(func, "__name__", "") == "__newobj__":
            #Added fix to allow transient
            cls = args[0]
            if not hasattr(cls, "__new__"):
                raise pickle.PicklingError(
                    "args[0] from __newobj__ args has no __new__")
            if obj is not None and cls is not obj.__class__:
                raise pickle.PicklingError(
                    "args[0] from __newobj__ args has the wrong class")
            args = args[1:]
            save(cls)

            #Don't pickle transient entries
            if hasattr(obj, '__transient__'):
                transient = obj.__transient__
                state = state.copy()

                for k in list(state.keys()):
                    if k in transient:
                        del state[k]

            save(args)
            write(pickle.NEWOBJ)
        else:
            save(func)
            save(args)
            write(pickle.REDUCE)

        if obj is not None:
            self.memoize(obj)

        # More new special cases (that work with older protocols as
        # well): when __reduce__ returns a tuple with 4 or 5 items,
        # the 4th and 5th item should be iterators that provide list
        # items and dict items (as (key, value) tuples), or None.

        if listitems is not None:
            self._batch_appends(listitems)

        if dictitems is not None:
            self._batch_setitems(dictitems)

        if state is not None:
            #print 'obj %s has state %s' % (obj, state)
            save(state)
            write(pickle.BUILD)


    def save_xrange(self, obj):
        """Save an xrange object in python 2.5
        Python 2.6 supports this natively
        """
        range_params = xrange_params(obj)
        self.save_reduce(_build_xrange,range_params)

    #python2.6+ supports xrange pickling. some py2.5 extensions might as well.  We just test it
    try:
        xrange(0).__reduce__()
    except TypeError: #can't pickle -- use PiCloud pickler
        dispatch[xrange] = save_xrange

    def save_partial(self, obj):
        """Partial objects do not serialize correctly in python2.x -- this fixes the bugs"""
        self.save_reduce(_genpartial, (obj.func, obj.args, obj.keywords))

    if sys.version_info < (2,7): #2.7 supports partial pickling
        dispatch[partial] = save_partial


    def save_file(self, obj):
        """Save a file"""
        import StringIO as pystringIO #we can't use cStringIO as it lacks the name attribute
        from ..transport.adapter import SerializingAdapter

        if not hasattr(obj, 'name') or  not hasattr(obj, 'mode'):
            raise pickle.PicklingError("Cannot pickle files that do not map to an actual file")
        if obj.name == '<stdout>':
            return self.save_reduce(getattr, (sys,'stdout'), obj=obj)
        if obj.name == '<stderr>':
            return self.save_reduce(getattr, (sys,'stderr'), obj=obj)
        if obj.name == '<stdin>':
            raise pickle.PicklingError("Cannot pickle standard input")
        if  hasattr(obj, 'isatty') and obj.isatty():
            raise pickle.PicklingError("Cannot pickle files that map to tty objects")
        if 'r' not in obj.mode:
            raise pickle.PicklingError("Cannot pickle files that are not opened for reading")
        name = obj.name
        try:
            fsize = os.stat(name).st_size
        except OSError:
            raise pickle.PicklingError("Cannot pickle file %s as it cannot be stat" % name)

        if obj.closed:
            #create an empty closed string io
            retval = pystringIO.StringIO("")
            retval.close()
        elif not fsize: #empty file
            retval = pystringIO.StringIO("")
            try:
                tmpfile = file(name)
                tst = tmpfile.read(1)
            except IOError:
                raise pickle.PicklingError("Cannot pickle file %s as it cannot be read" % name)
            tmpfile.close()
            if tst != '':
                raise pickle.PicklingError("Cannot pickle file %s as it does not appear to map to a physical, real file" % name)
        elif fsize > SerializingAdapter.max_transmit_data:
            raise pickle.PicklingError("Cannot pickle file %s as it exceeds cloudconf.py's max_transmit_data of %d" %
                                       (name,SerializingAdapter.max_transmit_data))
        else:
            try:
                tmpfile = file(name)
                contents = tmpfile.read(SerializingAdapter.max_transmit_data)
                tmpfile.close()
            except IOError:
                raise pickle.PicklingError("Cannot pickle file %s as it cannot be read" % name)
            retval = pystringIO.StringIO(contents)
            curloc = obj.tell()
            retval.seek(curloc)

        retval.name = name
        self.save(retval)  #save stringIO
        self.memoize(obj)

    dispatch[file] = save_file
    """Special functions for Add-on libraries"""

    def inject_numpy(self):
        numpy = sys.modules.get('numpy')
        if not numpy or not hasattr(numpy, 'ufunc'):
            return
        self.dispatch[numpy.ufunc] = self.__class__.save_ufunc

    numpy_tst_mods = ['numpy', 'scipy.special']
    def save_ufunc(self, obj):
        """Hack function for saving numpy ufunc objects"""
        name = obj.__name__
        for tst_mod_name in self.numpy_tst_mods:
            tst_mod = sys.modules.get(tst_mod_name, None)
            if tst_mod:
                if name in tst_mod.__dict__:
                    self.save_reduce(_getobject, (tst_mod_name, name))
                    return
        raise pickle.PicklingError('cannot save %s. Cannot resolve what module it is defined in' % str(obj))

    def inject_timeseries(self):
        """Handle bugs with pickling scikits timeseries"""
        tseries = sys.modules.get('scikits.timeseries.tseries')
        if not tseries or not hasattr(tseries, 'Timeseries'):
            return
        self.dispatch[tseries.Timeseries] = self.__class__.save_timeseries

    def save_timeseries(self, obj):
        import scikits.timeseries.tseries as ts

        func, reduce_args, state = obj.__reduce__()
        if func != ts._tsreconstruct:
            raise pickle.PicklingError('timeseries using unexpected reconstruction function %s' % str(func))
        state = (1,
                         obj.shape,
                         obj.dtype,
                         obj.flags.fnc,
                         obj._data.tostring(),
                         ts.getmaskarray(obj).tostring(),
                         obj._fill_value,
                         obj._dates.shape,
                         obj._dates.__array__().tostring(),
                         obj._dates.dtype, #added -- preserve type
                         obj.freq,
                         obj._optinfo,
                         )
        return self.save_reduce(_genTimeSeries, (reduce_args, state))

    def inject_email(self):
        """Block email LazyImporters from being saved"""
        email = sys.modules.get('email')
        if not email:
            return
        self.dispatch[email.LazyImporter] = self.__class__.save_unsupported

    def inject_addons(self):
        """Plug in system. Register additional pickling functions if modules already loaded"""
        self.inject_numpy()
        self.inject_timeseries()
        self.inject_email()

    """Python Imaging Library"""
    def save_image(self, obj):
        if not obj.im and obj.fp and 'r' in obj.fp.mode and obj.fp.name \
            and not obj.fp.closed and (not hasattr(obj, 'isatty') or not obj.isatty()):
            #if image not loaded yet -- lazy load
            self.save_reduce(_lazyloadImage,(obj.fp,), obj=obj)
        else:
            #image is loaded - just transmit it over
            self.save_reduce(_generateImage, (obj.size, obj.mode, obj.tostring()), obj=obj)

    """
    def memoize(self, obj):
        pickle.Pickler.memoize(self, obj)
        if printMemoization:
            print 'memoizing ' + str(obj)
    """



# Shorthands for legacy support

def dump(obj, file, protocol=2):
    CloudPickler(file, protocol).dump(obj)

def dumps(obj, protocol=2):
    file = StringIO()

    cp = CloudPickler(file,protocol)
    cp.dump(obj)

    #print 'cloud dumped', str(obj), str(cp.modules)

    return file.getvalue()


#hack for __import__ not working as desired
def subimport(name):
    __import__(name)
    return sys.modules[name]

#hack to load django settings:
def django_settings_load(name):
    modified_env = False

    if 'DJANGO_SETTINGS_MODULE' not in os.environ:
        os.environ['DJANGO_SETTINGS_MODULE'] = name # must set name first due to circular deps
        modified_env = True
    try:
        module = subimport(name)
    except Exception, i:
        print >> sys.stderr, 'Cloud not import django settings %s:' % (name)
        print_exec(sys.stderr)
        if modified_env:
            del os.environ['DJANGO_SETTINGS_MODULE']
    else:
        #add project directory to sys,path:
        if hasattr(module,'__file__'):
            dirname = os.path.split(module.__file__)[0] + '/'
            sys.path.append(dirname)

# restores function attributes
def _restore_attr(obj, attr):
    for key, val in attr.items():
        setattr(obj, key, val)
    return obj

def _get_module_builtins():
    return pickle.__builtins__

def print_exec(stream):
    ei = sys.exc_info()
    traceback.print_exception(ei[0], ei[1], ei[2], None, stream)

def _modules_to_main(modList):
    """Force every module in modList to be placed into main"""
    if not modList:
        return

    main = sys.modules['__main__']
    for modname in modList:
        if type(modname) is str:
            try:
                mod = __import__(modname)
            except Exception, i: #catch all...
                sys.stderr.write('warning: could not import %s\n.  Your function may unexpectedly error due to this import failing; \
A version mismatch is likely.  Specific error was:\n' % modname)
                print_exec(sys.stderr)
            else:
                setattr(main,mod.__name__, mod)
        else:
            #REVERSE COMPATIBILITY FOR CLOUD CLIENT 1.5 (WITH EPD)
            #In old version actual module was sent
            setattr(main,modname.__name__, modname)

#object generators:
def _build_xrange(start, step, len):
    """Built xrange explicitly"""
    return xrange(start, start + step*len, step)

def _genpartial(func, args, kwds):
    if not args:
        args = ()
    if not kwds:
        kwds = {}
    return partial(func, *args, **kwds)


def _fill_function(func, globals, defaults, closure, dict):
    """ Fills in the rest of function data into the skeleton function object
        that were created via _make_skel_func().
         """
    func.func_globals.update(globals)
    func.func_defaults = defaults
    func.func_dict = dict

    if len(closure) != len(func.func_closure):
        raise pickle.UnpicklingError("closure lengths don't match up")
    for i in range(len(closure)):
        _change_cell_value(func.func_closure[i], closure[i])

    return func

def _make_skel_func(code, num_closures, base_globals = None):
    """ Creates a skeleton function object that contains just the provided
        code and the correct number of cells in func_closure.  All other
        func attributes (e.g. func_globals) are empty.
    """
    #build closure (cells):
    if not ctypes:
        raise Exception('ctypes failed to import; cannot build function')

    cellnew = ctypes.pythonapi.PyCell_New
    cellnew.restype = ctypes.py_object
    cellnew.argtypes = (ctypes.py_object,)
    dummy_closure = tuple(map(lambda i: cellnew(None), range(num_closures)))

    if base_globals is None:
        base_globals = {}
    base_globals['__builtins__'] = __builtins__

    return types.FunctionType(code, base_globals,
                              None, None, dummy_closure)

# this piece of opaque code is needed below to modify 'cell' contents
cell_changer_code = new.code(
    1, 1, 2, 0,
    ''.join([
        chr(dis.opmap['LOAD_FAST']), '\x00\x00',
        chr(dis.opmap['DUP_TOP']),
        chr(dis.opmap['STORE_DEREF']), '\x00\x00',
        chr(dis.opmap['RETURN_VALUE'])
    ]),
    (), (), ('newval',), '<nowhere>', 'cell_changer', 1, '', ('c',), ()
)

def _change_cell_value(cell, newval):
    """ Changes the contents of 'cell' object to newval """
    return new.function(cell_changer_code, {}, None, (), (cell,))(newval)

"""Constructors for 3rd party libraries
Note: These can never be renamed due to client compatibility issues"""

def _getobject(modname, attribute):
    mod = __import__(modname, fromlist=[attribute])
    return mod.__dict__[attribute]

def _generateImage(size, mode, str_rep):
    """Generate image from string representation"""
    import Image
    i = Image.new(mode, size)
    i.fromstring(str_rep)
    return i

def _lazyloadImage(fp):
    import Image
    fp.seek(0)  #works in almost any case
    return Image.open(fp)

"""Timeseries"""
def _genTimeSeries(reduce_args, state):
    import scikits.timeseries.tseries as ts
    from numpy import ndarray
    from numpy.ma import MaskedArray


    time_series = ts._tsreconstruct(*reduce_args)

    #from setstate modified
    (ver, shp, typ, isf, raw, msk, flv, dsh, dtm, dtyp, frq, infodict) = state
    #print 'regenerating %s' % dtyp

    MaskedArray.__setstate__(time_series, (ver, shp, typ, isf, raw, msk, flv))
    _dates = time_series._dates
    #_dates.__setstate__((ver, dsh, typ, isf, dtm, frq))  #use remote typ
    ndarray.__setstate__(_dates,(dsh,dtyp, isf, dtm))
    _dates.freq = frq
    _dates._cachedinfo.update(dict(full=None, hasdups=None, steps=None,
                                   toobj=None, toord=None, tostr=None))
    # Update the _optinfo dictionary
    time_series._optinfo.update(infodict)
    return time_series




from __future__ import print_function
try:
    import gevent
    import gevent.monkey
    gevent.monkey.patch_all()
except ImportError:
    gevent = False
    pass
from argparse import ArgumentParser
import os
from os.path import abspath
import sys
import logging

import time
import urlparse

from redis.exceptions import ConnectionError

from simpleservices.redis import start_redis
from simpleservices.process import register_shutdown, ManagedProcess, close_all
from werkzeug.serving import run_with_reloader
from kitchensink.utils import parse_redis_connection
from kitchensink.rpc.server import make_app, run as runserver, register_rpc
from kitchensink.rpc import RPC
from kitchensink.data.datarpc import make_rpc as make_data_rpc
from kitchensink.admin import make_rpc as make_admin_rpc
from kitchensink.data import Catalog
import kitchensink.settings as settings
FORMAT = "%(created)f:%(name)s:%(message)s"
comments = \
"""
kitchen sink RPC Server

This command is used to start a single box RPC Server including a web gateway,
one Node, and N worker processes
"""
def parser():
    p = ArgumentParser(comments)
    p.add_argument('--redis-connection',
                   help="redis connection information, tcp://localhost:6379?db=9",
                   default="tcp://localhost:6379?db=9")
    p.add_argument('--node-url', help='url of node', default='http://localhost:6323/')
    p.add_argument('--no-redis',
                   help="do not start redis",
                   default=False,
                   action="store_true")
    p.add_argument('--node-port',
                   help="port for the main worker node of the RPC Server",
                   default=None)
    p.add_argument('--num-workers',
                   help="number of workers",
                   type=int,
                   default=1)
    p.add_argument('--queue',
                   help='queue to operate on',
                   action='append')
    p.add_argument('--reload',
                   help='reload when code has changed',
                   default=False,
                   action='store_true'
    )

    #mandatory
    p.add_argument('--datadir',
                   help='data directory',
                   default='tmp'
    )
    p.add_argument('--module',
                   help='(optional) module with rpc functions',
    )
    return p

def run_args(args):
    run(args.redis_connection, args.node_url,
        args.node_port,
        args.num_workers, args.no_redis, args.queue,
        args.module,
        args.datadir
    )

def run(redis_connection, node_url, node_port,
        num_workers, no_redis, queue, module, datadir):
    if not node_url.endswith("/"):
        node_url += "/"
    datadir = abspath(datadir)
    register_shutdown()
    redis_connection_info = parse_redis_connection(redis_connection)
    if node_port is None:
        node_port = int(urlparse.urlparse(node_url).netloc.split(":")[-1])
    print ("**port", node_port)
    pid_file = "ks-%s.pid" % node_port
    if not no_redis:
        print ("Starting redis on %s" % redis_connection)
        start_redis(pid_file, redis_connection_info['port'], os.getcwd())
        time.sleep(1)
    cmd = [sys.executable, '-m', 'kitchensink.scripts.start_worker',
           '--node-url', node_url,
           '--redis-connection', redis_connection,
           '--datadir', datadir,
    ]
    app = make_app(redis_connection_info, node_port, node_url, datadir)
    app.debug_log_format = FORMAT
    for c in range(10):
        try:
            result = settings.redis_conn.ping()
            break
        except ConnectionError:
            time.sleep(1.0)
    if queue is None:
        queue = ['default']
    for q in queue:
        cmd.extend(['--queue', q])
    for c in range(num_workers):
        ManagedProcess(cmd, 'worker-%s' % c, pid_file)
    data_rpc = make_data_rpc()
    register_rpc(data_rpc, 'data')
    default_rpc = RPC()
    register_rpc(default_rpc, 'default')
    admin = make_admin_rpc()
    register_rpc(admin, 'admin')
    if module:
        mod = __import__(module)
    runserver(gevent=gevent)
    #close_all()


def main():
    p = parser()
    args = p.parse_args()
    def helper():
        run_args(args)
    if args.reload:
        run_with_reloader(helper)
    else:
        helper()

if __name__ == "__main__":

    logging.basicConfig(level=logging.INFO, format=FORMAT)
    logging.getLogger("requests.packages.urllib3.connectionpool").setLevel(logging.WARNING)

    #logging.getLogger('rq.worker').level = logging.WARNING
    main()



# needs to be tested
from __future__ import print_function
import sys
if len(sys.argv)>1 and sys.argv[1] == 'develop':
    # Only import setuptools if we have to
    import site
    from os.path import dirname, abspath, join
    site_packages = site.getsitepackages()[0]
    fname = join(site_packages, "kitchensink.pth")
    path = abspath(dirname(__file__))
    with open(fname, "w+") as f:
        f.write(path)
    print("develop mode, wrote path (%s) to (%s)" % (path, fname))
    sys.exit()
from distutils.core import setup
import os
import sys
__version__ = (0, 2)
setup(
    name = 'kitchensinkRPC',
    version = '.'.join([str(x) for x in __version__]),
    packages = ['kitchensink',
                'kitchensink.admin',
                'kitchensink.api',
                'kitchensink.clients',
                'kitchensink.data',
                'kitchensink.rpc',
                'kitchensink.scripts',
                'kitchensink.serialization',
                'kitchensink.taskqueue',
                'kitchensink.testutils',
                'kitchensink.utils',
                ],

    url = 'http://github.com/hhuuggoo/kitchensink',
    description = 'Kitchen Sink',
    zip_safe=False,
    author='Hugo Shi',
    author_email='hugo.r.shi@gmail.com',
    license = 'New BSD',
)



import logging

from .decorators import remote

logger = logging.getLogger(__name__)

def fhead(obj, start=0, end=10):
    path = obj.local_path()
    results = []
    with open(path, 'r') as f:
        for idx, line in enumerate(f):
            results.append(line)
            if idx == end:
                break
    return results[start:]

def max(obj, col):
    return obj[col].obj().max()

def min(obj, col):
    return obj[col].obj().min()

def chunks(l, n):
    """ Yield successive n-sized chunks from l.
    """
    for i in range(0, len(l), n):
        yield l[i:i+n]

def workflow(c, func, data=[], force=False):
    compute = False
    if force:
        compute = True
    else:
        for prefix, threshold in data:
            num_objs = len(c.path_search(prefix))
            if num_objs < threshold:
                logger.info("%s objs found for prefix %s recomputing", num_objs, prefix)
                compute = True
                break
            else:
                logger.info("%s objs found for prefix %s no need to compute",
                            num_objs, prefix)
    if compute:
        logger.info("computing")
        for prefix, threshold in data:
            logger.info("clearing prefix %s", prefix)
            c.reducetree(prefix)
        func()



# todo use random ports
import tempfile
import sys
import os
import shutil
import time
import cPickle as pickle

import numpy as np
import requests
import pandas as pd
import redis
from simpleservices.redis import start_redis
from simpleservices.process import register_shutdown, ManagedProcess, close_all, close

from ..api import setup_client, client, du, do, dp, Client
from .. import settings

pid_file = 'ks-test.pid'
def setup_module():
    global dir1
    global dir2
    global dir3
    global redis_dir
    global rport
    global url1
    global url2
    global url3


    dir1 = tempfile.mkdtemp(prefix='ks-test')
    dir2 = tempfile.mkdtemp(prefix='ks-test')
    dir3 = tempfile.mkdtemp(prefix='ks-test')
    redis_dir = tempfile.mkdtemp(prefix='ks-test')

    port1 = 6323
    port2 = 6324
    port3 = 6325
    rport = 6326

    url1 = "http://localhost:%s/" % port1
    url2 = "http://localhost:%s/" % port2
    url3 = "http://localhost:%s/" % port3
    redis_url = 'tcp://localhost:%s?db=9' % rport
    start_redis(pid_file, rport, redis_dir, stdout=None, stderr=None)
    wait_redis_start(rport)
    cmd = ['python', '-m', 'kitchensink.scripts.start',
           '--datadir', dir1,
           '--no-redis',
           '--node-url', url1,
           '--num-workers', '1',
           '--redis-connection', redis_url]
    p = ManagedProcess(cmd, 'node1', pid_file)
    wait_ks_start(url1)
    cmd = ['python', '-m', 'kitchensink.scripts.start',
           '--datadir', dir2,
           '--no-redis',
           '--node-url', url2,
           '--num-workers', '1',
           '--redis-connection', redis_url]
    p = ManagedProcess(cmd, 'node2', pid_file)
    wait_ks_start(url2)
    cmd = ['python', '-m', 'kitchensink.scripts.start',
           '--datadir', dir3,
           '--no-redis',
           '--node-url', url3,
           '--num-workers', '1',
           '--redis-connection', redis_url]
    p = ManagedProcess(cmd, 'node3', pid_file)
    wait_ks_start(url3)
    # hack - wait for workers to sleep. necessary if
    # we want tests to accurately reflect performance
    time.sleep(1)

def teardown_module():
    close(pid_file, 'node1')
    close(pid_file, 'node2')
    close(pid_file, 'node3')
    wait_ks_gone(url1)
    wait_ks_gone(url2)
    wait_ks_gone(url3)
    close(pid_file, 'redis')
    wait_redis_gone(rport)
    close_all()
    shutil.rmtree(dir1)
    shutil.rmtree(dir2)
    shutil.rmtree(dir3)
    shutil.rmtree(redis_dir)

def wait_until(func, timeout=1.0, interval=0.01):
    st = time.time()
    while True:
        if func():
            return True
        if (time.time() - st) > timeout:
            return False
        time.sleep(interval)

def wait_redis_start(port):
    def helper():
        client = redis.Redis(port=port)
        try:
            return client.ping()
        except redis.ConnectionError:
            pass
    return wait_until(helper)

def wait_redis_gone(port):
    def helper():
        client = redis.Redis(port=port)
        try:
            client.ping()
            return False
        except redis.ConnectionError:
            return True
    return wait_until(helper)

def wait_ks_start(url):
    def helper():
        c = Client(url)
        return c.ping()
    return wait_until(helper)

def wait_ks_gone(url):
    def helper():
        c = Client(url)
        return c.ping() is None
    return wait_until(helper)

def dummy_func(x):
    return x

def remote_data_func(a, b):
    result = a.obj() + b.obj()
    result = do(obj=result)
    result.save()
    return result

def sleep_func():
    time.sleep(0.5)
    print 'sleep'
    time.sleep(0.5)
    print 'sleep'

def routing_func(*args):
    """test function which returns
    the host where data was executed
    """
    time.sleep(0.5)
    return settings.host_url

def remote_obj_func(obj):
    return obj.obj()

def remote_file(obj):
    path = obj.local_path()
    with open(path, 'rb') as f:
        return f.read()



"""module for rq subclasses
"""
import time
import random
import signal
import os
import sys
import traceback

from rq import Queue, Worker
from rq.job import Job, UNEVALUATED, Status, NoSuchJobError, UnpickleError
import rq.job
from rq.worker import StopRequested
from rq.utils import utcnow
from rq.compat import total_ordering, string_types, as_text
import dill

from ..serialization import serializer, deserializer
from ..utils import setup_loghandlers
import logging
logger = logging.getLogger(__name__)

def empty(result):
    """result of pop operation. dict of lists
    """
    for r in result.values():
        if len(r) != 0:
            return True
    return False

def nonblock_pop(connection, keys, timeout=5.0):
    for k in keys:
        msg = connection.lpop(k)
        if msg:
            msg = deserializer('json')(msg)
            return k, msg
    return None

def nonblock_popall(connection, keys):
    outputs = []
    for k in keys:
        msgs = connection.lrange(k, 0, -1)
        if msgs:
            connection.ltrim(k, len(msgs), -1)
            for m in msgs:
                outputs.append((k, deserializer('json')(m)))
    return outputs

def pop(connection, keys, timeout=5.0):
    msg = connection.blpop(keys, timeout=timeout)
    if msg is None:
        return None
    k, msg = msg
    msg = deserializer('json')(msg)
    return k, msg

def _grab_all_messages(connection, keys):
    return nonblock_popall(connection, keys)

def _block_and_grab_all_messages(connection, keys, timeout=5.0):
    messages = []
    msg = pop(connection, keys, timeout=timeout)
    if msg is None:
        return []
    else:
        messages = _grab_all_messages(connection, keys)
        return [msg] + messages

def pull_intermediate_results(connection, keys, timeout=5):
    """pull all messages off the queue.  We pull objects
    in a non blocking manner first.  If we something, we return that
    something. if we get nothing, then we do a blocking pop
    for timeout
    """
    keys = ["rq:job:"+ x + ":intermediate_results" for x in keys]
    messages = _grab_all_messages(connection, keys)
    #if timeout is 0, then we want instant return, do not do blocking call
    if not messages and timeout != 0:
        messages =  _block_and_grab_all_messages(connection, keys, timeout=timeout)
    messages = [(x[0].split(":")[2], x[1]) for x in messages]
    return messages

class KitchenSinkJob(Job):

    def set_status(self, status, pipeline=None):
        connection = pipeline if pipeline is not None else self.connection
        self._status = status
        connection.hset(self.key, 'status', self._status)

    @property
    def intermediate_results_key(self):
        return self.key + ":" + "intermediate_results"

    @classmethod
    def claim_key_for(cls, job_id):
        return cls.key_for(job_id) + ":claim"

    @classmethod
    def claim_for(cls, connection, job_id, queue_name):
        val = connection.setnx(cls.claim_key_for(job_id), queue_name)
        if val == 0:
            return False
        else:
            return True

    def push_intermediate_results(self, result, pipeline=None):
        connection = pipeline if pipeline is not None else self.connection
        msg = serializer('json')(result)
        connection.rpush(self.intermediate_results_key, msg)

    def cleanup(self, ttl=None, pipeline=None):
        super(KitchenSinkJob, self).cleanup(ttl=ttl, pipeline=pipeline)
        self.connection.expire(self.intermediate_results_key, ttl)

    def push_status(self, status=None, pipeline=None):
        connection = pipeline if pipeline is not None else self.connection
        if status is None:
            status = self.get_status()
        self.push_intermediate_results({'type' : 'status',
                                        'status' : status},
                                       pipeline=connection)
    def push_stdout(self, output):
        self.push_intermediate_results({'type' : 'stdout',
                                        'msg' : output})

    def pull_intermediate_results(self, timeout=5):
        """pull all messages off the queue.  We pull objects
        in a non blocking manner first.  If we something, we return that
        something. if we get nothing, then we do a blocking pop
        for timeout
        """
        messages = pull_intermediate_results(self.connection,
                                             [self.id],
                                             timeout=timeout)
        messages = [x[1] for x in messages]
        return messages

    # Job execution
    def perform(self):  # noqa
        """Invokes the job function with the job arguments."""
        rq.job._job_stack.push(self.id)
        try:
            self._result = self.func(*self.args, **self.kwargs)
            self.ended_at = utcnow()
        finally:
            assert self.id == rq.job._job_stack.pop()
        return self._result

class KitchenSinkRedisQueue(Queue):
    DEFAULT_TIMEOUT = 86400
    job_class = KitchenSinkJob
    def enqueue_job(self, job, set_meta_data=True):
        """Enqueues a job for delayed execution.

        If the `set_meta_data` argument is `True` (default), it will update
        the properties `origin` and `enqueued_at`.

        If Queue is instantiated with async=False, job is executed immediately.
        """
        # Add Queue key set
        self.connection.sadd(self.redis_queues_keys, self.key)
        if self._async:
            self.push_job_id(job.id)
        else:
            job.perform()
            job.save()
        return job

    @classmethod
    def dequeue_any(cls, queues, timeout, connection=None):
        while True:
            #exits on timeout error
            result = cls._dequeue_any(queues, timeout, connection=connection)
            if result is not None:
                return result

    @classmethod
    def _dequeue_any(cls, queues, timeout, connection=None):
        """Class method returning the job_class instance at the front of the given
        set of Queues, where the order of the queues is important.

        When all of the Queues are empty, depending on the `timeout` argument,
        either blocks execution of this function for the duration of the
        timeout or until new messages arrive on any of the queues, or returns
        None.

        See the documentation of cls.lpop for the interpretation of timeout.
        """
        queue_keys = [q.key for q in queues]
        result = cls.lpop(queue_keys, timeout, connection=connection)
        if result is None:
            return None
        queue_key, job_id = map(as_text, result)
        if connection is None:
            connection = self.connection
        """We now enqueue jobs on multiple queues, so when we pop them
        we have to try to claim it.  if we can't claim it, discard,
        that means another worker has it
        """
        if not cls.job_class.claim_for(connection, job_id, queue_key):
            return None
        queue = cls.from_queue_key(queue_key, connection=connection)
        try:
            job = cls.job_class.fetch(job_id, connection=connection)
        except NoSuchJobError:
            # Silently pass on jobs that don't exist (anymore),
            # and continue by reinvoking the same function recursively
            return None
        except UnpickleError as e:
            # Attach queue information on the exception for improved error
            # reporting
            e.job_id = job_id
            e.queue = queue
            raise e
        return job, queue

class KitchenSinkWorker(Worker):
    job_class = KitchenSinkJob
    queue_class = KitchenSinkRedisQueue
    def work(self, burst=False):
        setup_loghandlers()
        super(KitchenSinkWorker, self).work(burst=burst)

    def heartbeat(self, timeout=0, pipeline=None):
        """Specifies a new worker timeout, typically by extending the
        expiration time of the worker, effectively making this a "heartbeat"
        to not expire the worker until the timeout passes.

        The next heartbeat should come before this time, or the worker will
        die (at least from the monitoring dashboards).

        The effective timeout can never be shorter than default_worker_ttl,
        only larger.
        """
        connection = pipeline if pipeline is not None else self.connection
        timeout = max(timeout, self.default_worker_ttl)
        connection.expire(self.key, timeout)
        self.log.debug('Sent heartbeat to prevent worker timeout. '
                       'Next one should arrive within {0} seconds.'.format(timeout))

    def perform_job(self, job):
        """Performs the actual work of a job.  Will/should only be called
        inside the work horse's process.
        """
        with self.connection._pipeline() as pipeline:
            self.heartbeat((job.timeout or 180) + 60, pipeline=pipeline)
            self.set_state('busy', pipeline=pipeline)
            self.set_current_job_id(job.id, pipeline=pipeline)
            job.set_status(Status.STARTED, pipeline=pipeline)
            job.push_status(Status.STARTED, pipeline=pipeline)
            pipeline.execute()
        with self.connection._pipeline() as pipeline:
            try:
                with self.death_penalty_class(job.timeout or self.queue_class.DEFAULT_TIMEOUT):
                    rv = job.perform()
                # Pickle the result in the same try-except block since we need to
                # use the same exc handling when pickling fails
                job._result = rv
                job._status = Status.FINISHED
                self.set_current_job_id(None, pipeline=pipeline)
                result_ttl = job.get_ttl(self.default_result_ttl)
                if result_ttl != 0:
                    job.save(pipeline=pipeline)
                job.cleanup(result_ttl, pipeline=pipeline)
                job.push_status(status=Status.FINISHED, pipeline=pipeline)
                pipeline.execute()

            except Exception:
                # Use the public setter here, to immediately update Redis
                self.handle_exception(job, *sys.exc_info())
                job.set_status(Status.FAILED)
                job.push_status()
                return False

        if rv is None:
            self.log.info('Job OK')
        else:
            self.log.info('Job OK, result = %s', rv)

        if result_ttl == 0:
            self.log.info('Result discarded immediately.')
        elif result_ttl > 0:
            self.log.info('Result is kept for %d seconds.' % result_ttl)
        else:
            self.log.warning('Result will never expire, clean up result key manually.')

        return True


def current_job_id():
    """Returns the Job instance that is currently being executed.  If this
    function is invoked from outside a job context, None is returned.
    """
    job_id = rq.job._job_stack.top
    if job_id is None:
        return None
    else:
        return job_id
